"""
This module provides the `Data` class, which offers
methods to work with nested data structures.
"""

from .base.types import DataStructureTypes, IndexIterableTypes, DataStructure, IndexIterable

from .format_codes import FormatCodes
from .string import String
from .regex import Regex

from typing import Optional, Literal, Final, Any, cast
import base64 as _base64
import math as _math
import re as _re


_DEFAULT_SYNTAX_HL: Final[dict[str, tuple[str, str]]] = {
    "str": ("[br:blue]", "[_c]"),
    "number": ("[br:magenta]", "[_c]"),
    "literal": ("[magenta]", "[_c]"),
    "type": ("[i|green]", "[_i|_c]"),
    "punctuation": ("[br:black]", "[_c]"),
}
"""Default syntax highlighting styles for data structure rendering."""


class Data:
    """This class includes methods to work with nested data structures (dictionaries and lists)."""

    @classmethod
    def serialize_bytes(cls, data: bytes | bytearray) -> dict[str, str]:
        """Converts bytes or bytearray to a JSON-compatible format (dictionary) with explicit keys.\n
        ----------------------------------------------------------------------------------------------
        - `data` -⠀the bytes or bytearray to serialize"""
        key = "bytearray" if isinstance(data, bytearray) else "bytes"

        try:
            return {key: cast(bytes | bytearray, data).decode("utf-8"), "encoding": "utf-8"}
        except UnicodeDecodeError:
            pass

        return {key: _base64.b64encode(data).decode("utf-8"), "encoding": "base64"}

    @classmethod
    def deserialize_bytes(cls, obj: dict[str, str]) -> bytes | bytearray:
        """Tries to converts a JSON-compatible bytes/bytearray format (dictionary) back to its original type.\n
        --------------------------------------------------------------------------------------------------------
        - `obj` -⠀the dictionary to deserialize\n
        --------------------------------------------------------------------------------------------------------
        If the serialized object was created with `Data.serialize_bytes()`, it will work.
        If it fails to decode the data, it will raise a `ValueError`."""
        for key in ("bytes", "bytearray"):
            if key in obj and "encoding" in obj:
                if obj["encoding"] == "utf-8":
                    data = obj[key].encode("utf-8")
                elif obj["encoding"] == "base64":
                    data = _base64.b64decode(obj[key].encode("utf-8"))
                else:
                    raise ValueError(f"Unknown encoding method '{obj['encoding']}'")

                return bytearray(data) if key == "bytearray" else data

        raise ValueError(f"Invalid serialized data:\n  {obj}")

    @classmethod
    def chars_count(cls, data: DataStructure) -> int:
        """The sum of all the characters amount including the keys in dictionaries.\n
        ------------------------------------------------------------------------------
        - `data` -⠀the data structure to count the characters from"""
        chars_count = 0

        if isinstance(data, dict):
            for k, v in data.items():
                chars_count += len(str(k)) + (cls.chars_count(v) if isinstance(v, DataStructureTypes) else len(str(v)))

        elif isinstance(data, IndexIterableTypes):
            for item in data:
                chars_count += cls.chars_count(item) if isinstance(item, DataStructureTypes) else len(str(item))

        return chars_count

    @classmethod
    def strip(cls, data: DataStructure) -> DataStructure:
        """Removes leading and trailing whitespaces from the data structure's items.\n
        -------------------------------------------------------------------------------
        - `data` -⠀the data structure to strip the items from"""
        if isinstance(data, dict):
            return {k.strip(): cls.strip(v) if isinstance(v, DataStructureTypes) else v.strip() for k, v in data.items()}

        if isinstance(data, IndexIterableTypes):
            return type(data)(cls.strip(item) if isinstance(item, DataStructureTypes) else item.strip() for item in data)

        raise TypeError(f"Unsupported data structure type: {type(data)}")

    @classmethod
    def remove_empty_items(cls, data: DataStructure, spaces_are_empty: bool = False) -> DataStructure:
        """Removes empty items from the data structure.\n
        ---------------------------------------------------------------------------------
        - `data` -⠀the data structure to remove empty items from.
        - `spaces_are_empty` -⠀if true, it will count items with only spaces as empty"""
        if isinstance(data, dict):
            return {
                k: (v if not isinstance(v, DataStructureTypes) else cls.remove_empty_items(v, spaces_are_empty))
                for k, v in data.items() if not String.is_empty(v, spaces_are_empty)
            }

        if isinstance(data, IndexIterableTypes):
            return type(data)(
                item for item in
                (
                    (item if not isinstance(item, DataStructureTypes) else cls.remove_empty_items(item, spaces_are_empty)) \
                    for item in data if not (isinstance(item, (str, type(None))) and String.is_empty(item, spaces_are_empty))
                )
                if item not in ([], (), {}, set(), frozenset())
            )

        raise TypeError(f"Unsupported data structure type: {type(data)}")

    @classmethod
    def remove_duplicates(cls, data: DataStructure) -> DataStructure:
        """Removes all duplicates from the data structure.\n
        -----------------------------------------------------------
        - `data` -⠀the data structure to remove duplicates from"""
        if isinstance(data, dict):
            return {k: cls.remove_duplicates(v) if isinstance(v, DataStructureTypes) else v for k, v in data.items()}

        if isinstance(data, (list, tuple)):
            result: list[Any] = []
            for item in data:
                processed_item = cls.remove_duplicates(item) if isinstance(item, DataStructureTypes) else item
                is_duplicate: bool = False

                for existing_item in result:
                    if processed_item == existing_item:
                        is_duplicate = True
                        break

                if not is_duplicate:
                    result.append(processed_item)

            return type(data)(result)

        if isinstance(data, (set, frozenset)):
            processed_elements = set()
            for item in data:
                processed_item = cls.remove_duplicates(item) if isinstance(item, DataStructureTypes) else item
                processed_elements.add(processed_item)
            return type(data)(processed_elements)

        raise TypeError(f"Unsupported data structure type: {type(data)}")

    @classmethod
    def remove_comments(
        cls,
        data: DataStructure,
        comment_start: str = ">>",
        comment_end: str = "<<",
        comment_sep: str = "",
    ) -> DataStructure:
        """Remove comments from a list, tuple or dictionary.\n
        ---------------------------------------------------------------------------------------------------------------
        - `data` -⠀list, tuple or dictionary, where the comments should get removed from
        - `comment_start` -⠀the string that marks the start of a comment inside `data`
        - `comment_end` -⠀the string that marks the end of a comment inside `data`
        - `comment_sep` -⠀the string with which a comment will be replaced, if it is in the middle of a value\n
        ---------------------------------------------------------------------------------------------------------------
        #### Examples:
        ```python
        data = {
            "key1": [
                ">> COMMENT IN THE BEGINNING OF THE STRING <<  value1",
                "value2  >> COMMENT IN THE END OF THE STRING",
                "val>> COMMENT IN THE MIDDLE OF THE STRING <<ue3",
                ">> FULL VALUE IS A COMMENT  value4",
            ],
            ">> FULL KEY + ALL ITS VALUES ARE A COMMENT  key2": [
                "value",
                "value",
                "value",
            ],
            "key3": ">> ALL THE KEYS VALUES ARE COMMENTS  value",
        }

        processed_data = Data.remove_comments(
            data,
            comment_start=">>",
            comment_end="<<",
            comment_sep="__",
        )
        ```\n
        ---------------------------------------------------------------------------------------------------------------
        For this example, `processed_data` will be:
        ```python
        {
            "key1": [
                "value1",
                "value2",
                "val__ue3",
            ],
            "key3": None,
        }
        ```\n
        - For `key1`, all the comments will just be removed, except at `value3` and `value4`:
          * `value3` The comment is removed and the parts left and right are joined through `comment_sep`.
          * `value4` The whole value is removed, since the whole value was a comment.
        - For `key2`, the key, including its whole values will be removed.
        - For `key3`, since all its values are just comments, the key will still exist, but with a value of `None`."""
        if len(comment_start) == 0:
            raise ValueError("The 'comment_start' parameter string must not be empty.")

        return _DataRemoveCommentsHelper(
            data=data,
            comment_start=comment_start,
            comment_end=comment_end,
            comment_sep=comment_sep,
        )()

    @classmethod
    def is_equal(
        cls,
        data1: DataStructure,
        data2: DataStructure,
        ignore_paths: str | list[str] = "",
        path_sep: str = "->",
        comment_start: str = ">>",
        comment_end: str = "<<",
    ) -> bool:
        """Compares two structures and returns `True` if they are equal and `False` otherwise.\n
        ⇾ Will not detect, if a key-name has changed, only if removed or added.\n
        ------------------------------------------------------------------------------------------------
        - `data1` -⠀the first data structure to compare
        - `data2` -⠀the second data structure to compare
        - `ignore_paths` -⠀a path or list of paths to key/s and item/s to ignore during comparison:<br>
          Comments are not ignored when comparing. `comment_start` and `comment_end` are only used
          to correctly recognize the keys in the `ignore_paths`.
        - `path_sep` -⠀the separator between the keys/indexes in the `ignore_paths`
        - `comment_start` -⠀the string that marks the start of a comment inside `data1` and `data2`
        - `comment_end` -⠀the string that marks the end of a comment inside `data1` and `data2`\n
        ------------------------------------------------------------------------------------------------
        The paths from `ignore_paths` and the `path_sep` parameter work exactly the same way as for
        the method `Data.get_path_id()`. See its documentation for more details."""
        if len(path_sep) == 0:
            raise ValueError("The 'path_sep' parameter string must not be empty.")

        if isinstance(ignore_paths, str):
            ignore_paths = [ignore_paths]

        return cls._compare_nested(
            data1=cls.remove_comments(data1, comment_start, comment_end),
            data2=cls.remove_comments(data2, comment_start, comment_end),
            ignore_paths=[str(path).split(path_sep) for path in ignore_paths if path],
        )

    @classmethod
    def get_path_id(
        cls,
        data: DataStructure,
        value_paths: str | list[str],
        path_sep: str = "->",
        comment_start: str = ">>",
        comment_end: str = "<<",
        ignore_not_found: bool = False,
    ) -> Optional[str | list[Optional[str]]]:
        """Generates a unique ID based on the path to a specific value within a nested data structure.\n
        --------------------------------------------------------------------------------------------------
        -`data` -⠀the list, tuple, or dictionary, which the id should be generated for
        - `value_paths` -⠀a path or list of paths to the value/s to generate the id for (explained below)
        - `path_sep` -⠀the separator between the keys/indexes in the `value_paths`
        - `comment_start` -⠀the string that marks the start of a comment inside `data`
        - `comment_end` -⠀the string that marks the end of a comment inside `data`
        - `ignore_not_found` -⠀if true, the function will return `None` if the value is not found
          instead of raising an error\n
        --------------------------------------------------------------------------------------------------
        The param `value_path` is a sort of path (or a list of paths) to the value/s to be updated.
        #### In this example:
        ```python
        {
            "healthy": {
                "fruit": ["apples", "bananas", "oranges"],
                "vegetables": ["carrots", "broccoli", "celery"]
            }
        }
        ```
        … if you want to change the value of `"apples"` to `"strawberries"`, the value path would be
        `healthy->fruit->apples` or if you don't know that the value is `"apples"` you can also use the
        index of the value, so `healthy->fruit->0`."""
        if len(path_sep) == 0:
            raise ValueError("The 'path_sep' parameter string must not be empty.")

        data = cls.remove_comments(data, comment_start, comment_end)
        if isinstance(value_paths, str):
            return _DataGetPathIdHelper(value_paths, path_sep, data, ignore_not_found)()

        results = [_DataGetPathIdHelper(path, path_sep, data, ignore_not_found)() for path in value_paths]
        return results if len(results) > 1 else results[0] if results else None

    @classmethod
    def get_value_by_path_id(cls, data: DataStructure, path_id: str, get_key: bool = False) -> Any:
        """Retrieves the value from `data` using the provided `path_id`, as long as the data structure
        hasn't changed since creating the path ID.\n
        --------------------------------------------------------------------------------------------------
        - `data` -⠀the list, tuple, or dictionary to retrieve the value from
        - `path_id` -⠀the path ID to the value to retrieve, created before using `Data.get_path_id()`
        - `get_key` -⠀if true and the final item is in a dict, it returns the key instead of the value"""
        parent: Optional[DataStructure] = None
        path = cls._sep_path_id(path_id)
        current_data: Any = data

        for i, path_idx in enumerate(path):
            if isinstance(current_data, dict):
                keys = list(current_data.keys())
                if i == len(path) - 1 and get_key:
                    return keys[path_idx]
                parent = current_data
                current_data = current_data[keys[path_idx]]

            elif isinstance(current_data, IndexIterableTypes):
                if i == len(path) - 1 and get_key:
                    if parent is None or not isinstance(parent, dict):
                        raise ValueError(f"Cannot get key from a non-dict parent at path '{path[:i+1]}'")
                    return next(key for key, value in parent.items() if value is current_data)
                parent = current_data
                current_data = list(current_data)[path_idx]  # CONVERT TO LIST FOR INDEXING

            else:
                raise TypeError(f"Unsupported type '{type(current_data)}' at path '{path[:i+1]}'")

        return current_data

    @classmethod
    def set_value_by_path_id(cls, data: DataStructure, update_values: dict[str, Any]) -> DataStructure:
        """Updates the value/s from `update_values` in the `data`, as long as the data structure
        hasn't changed since creating the path ID to that value.\n
        -----------------------------------------------------------------------------------------
        - `data` -⠀the list, tuple, or dictionary to update the value/s in
        - `update_values` -⠀a dictionary where keys are path IDs and values are the new values
          to insert, for example:
          ```python
        { "1>012": "new value", "1>31": ["new value 1", "new value 2"], … }
          ```
          The path IDs should have been created using `Data.get_path_id()`."""
        if not (valid_update_values := [(path_id, new_val) for path_id, new_val in update_values.items()]):
            raise ValueError(f"No valid 'update_values' found in dictionary:\n{update_values!r}")

        for path_id, new_val in valid_update_values:
            data = cls._set_nested_val(data, id_path=cls._sep_path_id(path_id), value=new_val)

        return data

    @classmethod
    def render(
        cls,
        data: DataStructure,
        indent: int = 4,
        compactness: Literal[0, 1, 2] = 1,
        max_width: int = 127,
        sep: str = ", ",
        as_json: bool = False,
        syntax_highlighting: dict[str, str] | bool = False,
    ) -> str:
        """Get nicely formatted data structure-strings.\n
        ---------------------------------------------------------------------------------------------------------------
        - `data` -⠀the data structure to format
        - `indent` -⠀the amount of spaces to use for indentation
        - `compactness` -⠀the level of compactness for the output (explained below – section 1)
        - `max_width` -⠀the maximum width of a line before expanding (only used if `compactness` is `1`)
        - `sep` -⠀the separator between items in the data structure
        - `as_json` -⠀if true, the output will be in valid JSON format
        - `syntax_highlighting` -⠀a dictionary defining the syntax highlighting styles (explained below – section 2)
          or `True` to apply default syntax highlighting styles or `False`/`None` to disable syntax highlighting\n
        ---------------------------------------------------------------------------------------------------------------
        There are three different levels of `compactness`:
        - `0` expands everything possible
        - `1` only expands if there's other lists, tuples or dicts inside of data or,
          if the data's content is longer than `max_width`
        - `2` keeps everything collapsed (all on one line)\n
        ---------------------------------------------------------------------------------------------------------------
        The `syntax_highlighting` dictionary has 5 keys for each part of the data.<br>
        The key's values are the formatting codes to apply to this data part.<br>
        The formatting can be changed by simply adding the key with the new value
        inside the `syntax_highlighting` dictionary.\n
        The keys with their default values are:
        - `str: "br:blue"`
        - `number: "br:magenta"`
        - `literal: "magenta"`
        - `type: "i|green"`
        - `punctuation: "br:black"`\n
        ---------------------------------------------------------------------------------------------------------------
        For more detailed information about formatting codes, see the `format_codes` module documentation."""
        if indent < 0:
            raise ValueError("The 'indent' parameter must be a non-negative integer.")
        if max_width <= 0:
            raise ValueError("The 'max_width' parameter must be a positive integer.")

        return _DataRenderHelper(
            cls,
            data=data,
            indent=indent,
            compactness=compactness,
            max_width=max_width,
            sep=sep,
            as_json=as_json,
            syntax_highlighting=syntax_highlighting,
        )()

    @classmethod
    def print(
        cls,
        data: DataStructure,
        indent: int = 4,
        compactness: Literal[0, 1, 2] = 1,
        max_width: int = 127,
        sep: str = ", ",
        end: str = "\n",
        as_json: bool = False,
        syntax_highlighting: dict[str, str] | bool = {},
    ) -> None:
        """Print nicely formatted data structures.\n
        ---------------------------------------------------------------------------------------------------------------
        - `data` -⠀the data structure to format and print
        - `indent` -⠀the amount of spaces to use for indentation
        - `compactness` -⠀the level of compactness for the output (explained below – section 1)
        - `max_width` -⠀the maximum width of a line before expanding (only used if `compactness` is `1`)
        - `sep` -⠀the separator between items in the data structure
        - `end` -⠀the string appended after the last value, default a newline `\\n`
        - `as_json` -⠀if true, the output will be in valid JSON format
        - `syntax_highlighting` -⠀a dictionary defining the syntax highlighting styles (explained below – section 2)\n
        ---------------------------------------------------------------------------------------------------------------
        There are three different levels of `compactness`:
        - `0` expands everything possible
        - `1` only expands if there's other lists, tuples or dicts inside of data or,
          if the data's content is longer than `max_width`
        - `2` keeps everything collapsed (all on one line)\n
        ---------------------------------------------------------------------------------------------------------------
        The `syntax_highlighting` parameter is a dictionary with 5 keys for each part of the data.<br>
        The key's values are the formatting codes to apply to this data part.<br>
        The formatting can be changed by simply adding the key with the new value inside the
        `syntax_highlighting` dictionary.\n
        The keys with their default values are:
        - `str: "br:blue"`
        - `number: "br:magenta"`
        - `literal: "magenta"`
        - `type: "i|green"`
        - `punctuation: "br:black"`\n
        For no syntax highlighting, set `syntax_highlighting` to `False` or `None`.\n
        ---------------------------------------------------------------------------------------------------------------
        For more detailed information about formatting codes, see the `format_codes` module documentation."""
        FormatCodes.print(
            cls.render(
                data=data,
                indent=indent,
                compactness=compactness,
                max_width=max_width,
                sep=sep,
                as_json=as_json,
                syntax_highlighting=syntax_highlighting,
            ),
            end=end,
        )

    @classmethod
    def _compare_nested(
        cls,
        data1: Any,
        data2: Any,
        ignore_paths: list[list[str]],
        current_path: list[str] = [],
    ) -> bool:
        if any(current_path == path[:len(current_path)] for path in ignore_paths):
            return True

        if type(data1) is not type(data2):
            return False

        if isinstance(data1, dict) and isinstance(data2, dict):
            if set(data1.keys()) != set(data2.keys()):
                return False
            return all(cls._compare_nested( \
                data1=data1[key],
                data2=data2[key],
                ignore_paths=ignore_paths,
                current_path=current_path + [key],
            ) for key in data1)

        elif isinstance(data1, (list, tuple)):
            if len(data1) != len(data2):
                return False
            return all(cls._compare_nested( \
                data1=item1,
                data2=item2,
                ignore_paths=ignore_paths,
                current_path=current_path + [str(i)],
            ) for i, (item1, item2) in enumerate(zip(data1, data2)))

        elif isinstance(data1, (set, frozenset)):
            return data1 == data2

        return data1 == data2

    @staticmethod
    def _sep_path_id(path_id: str) -> list[int]:
        """Internal method to separate a path-ID string into its ID parts as a list of integers."""
        if len(split_id := path_id.split(">")) == 2:
            id_part_len, path_id_parts = split_id

            if (id_part_len.isdigit() and path_id_parts.isdigit()):
                id_part_len_int = int(id_part_len)

                if id_part_len_int > 0 and (len(path_id_parts) % id_part_len_int == 0):
                    return [int(path_id_parts[i:i + id_part_len_int]) for i in range(0, len(path_id_parts), id_part_len_int)]

        raise ValueError(f"Path ID '{path_id}' is an invalid format.")

    @classmethod
    def _set_nested_val(cls, data: DataStructure, id_path: list[int], value: Any) -> Any:
        """Internal method to set a value in a nested data structure based on the provided ID path."""
        current_data: Any = data

        if len(id_path) == 1:
            if isinstance(current_data, dict):
                keys, data_dict = list(current_data.keys()), dict(current_data)
                data_dict[keys[id_path[0]]] = value
                return data_dict
            elif isinstance(current_data, IndexIterableTypes):
                was_t, data_list = type(current_data), list(current_data)
                data_list[id_path[0]] = value
                return was_t(data_list)

        else:
            if isinstance(current_data, dict):
                keys, data_dict = list(current_data.keys()), dict(current_data)
                data_dict[keys[id_path[0]]] = cls._set_nested_val(data_dict[keys[id_path[0]]], id_path[1:], value)
                return data_dict
            elif isinstance(current_data, IndexIterableTypes):
                was_t, data_list = type(current_data), list(current_data)
                data_list[id_path[0]] = cls._set_nested_val(data_list[id_path[0]], id_path[1:], value)
                return was_t(data_list)

        return current_data


class _DataRemoveCommentsHelper:
    """Internal, callable helper class to remove all comments from nested data structures."""

    def __init__(self, data: DataStructure, comment_start: str, comment_end: str, comment_sep: str):
        self.data = data
        self.comment_start = comment_start
        self.comment_end = comment_end
        self.comment_sep = comment_sep

        self.pattern = _re.compile(Regex._clean( \
            rf"""^(
                (?:(?!{_re.escape(comment_start)}).)*
            )
            {_re.escape(comment_start)}
            (?:(?:(?!{_re.escape(comment_end)}).)*)
            (?:{_re.escape(comment_end)})?
            (.*?)$"""
        )) if len(comment_end) > 0 else None

    def __call__(self) -> DataStructure:
        return self.remove_nested_comments(self.data)

    def remove_nested_comments(self, item: Any) -> Any:
        if isinstance(item, dict):
            return {
                key: val
                for key, val in ( \
                    (self.remove_nested_comments(k), self.remove_nested_comments(v)) for k, v in item.items()
                ) if key is not None
            }

        if isinstance(item, IndexIterableTypes):
            processed = (v for v in map(self.remove_nested_comments, item) if v is not None)
            return type(item)(processed)

        if isinstance(item, str):
            if self.pattern:
                if (match := self.pattern.match(item)):
                    start, end = match.group(1).strip(), match.group(2).strip()
                    return f"{start}{self.comment_sep if start and end else ''}{end}" or None
                return item.strip() or None
            else:
                return None if item.lstrip().startswith(self.comment_start) else item.strip() or None

        return item


class _DataGetPathIdHelper:
    """Internal, callable helper class to process a data path and generate its unique path ID."""

    def __init__(self, path: str, path_sep: str, data_obj: DataStructure, ignore_not_found: bool):
        self.keys = path.split(path_sep)
        self.data_obj = data_obj
        self.ignore_not_found = ignore_not_found

        self.path_ids: list[str] = []
        self.max_id_length = 0
        self.current_data: Any = data_obj

    def __call__(self) -> Optional[str]:
        for key in self.keys:
            if not self.process_key(key):
                break

        if not self.path_ids:
            return None
        return f"{self.max_id_length}>{''.join(id.zfill(self.max_id_length) for id in self.path_ids)}"

    def process_key(self, key: str) -> bool:
        """Process a single key and update `path_ids`. Returns `False` if processing should stop."""
        idx: Optional[int] = None

        if isinstance(self.current_data, dict):
            if (idx := self.process_dict_key(key)) is None:
                return False
        elif isinstance(self.current_data, IndexIterableTypes):
            if (idx := self.process_iterable_key(key)) is None:
                return False
        else:
            return False

        self.path_ids.append(str(idx))
        self.max_id_length = max(self.max_id_length, len(str(idx)))
        return True

    def process_dict_key(self, key: str) -> Optional[int]:
        """Process a key for dictionary data. Returns the index or `None` if not found."""
        if key.isdigit():
            if self.ignore_not_found:
                return None
            raise TypeError(f"Key '{key}' is invalid for a dict type.")

        try:
            idx = list(self.current_data.keys()).index(key)
            self.current_data = self.current_data[key]
            return idx
        except (ValueError, KeyError):
            if self.ignore_not_found:
                return None
            raise KeyError(f"Key '{key}' not found in dict.")

    def process_iterable_key(self, key: str) -> Optional[int]:
        """Process a key for iterable data. Returns the index or `None` if not found."""
        try:
            idx = int(key)
            self.current_data = list(self.current_data)[idx]
            return idx
        except ValueError:
            try:
                idx = list(self.current_data).index(key)
                self.current_data = list(self.current_data)[idx]
                return idx
            except ValueError:
                if self.ignore_not_found:
                    return None
                raise ValueError(f"Value '{key}' not found in '{type(self.current_data).__name__}'")


class _DataRenderHelper:
    """Internal, callable helper class to format data structures as strings."""

    def __init__(
        self,
        cls: type[Data],
        data: DataStructure,
        indent: int,
        compactness: Literal[0, 1, 2],
        max_width: int,
        sep: str,
        as_json: bool,
        syntax_highlighting: dict[str, str] | bool,
    ):
        self.cls = cls
        self.data = data
        self.indent = indent
        self.compactness = compactness
        self.max_width = max_width
        self.as_json = as_json

        self.syntax_hl: dict[str, tuple[str, str]] = _DEFAULT_SYNTAX_HL.copy()
        self.do_syntax_hl = syntax_highlighting not in {None, False}

        if self.do_syntax_hl:
            if syntax_highlighting is True:
                syntax_highlighting = {}
            elif not isinstance(syntax_highlighting, dict):
                raise TypeError(f"Expected 'syntax_highlighting' to be a dict or bool. Got: {type(syntax_highlighting)}")

            self.syntax_hl.update({
                k: (f"[{v}]", "[_]") if k in self.syntax_hl and v not in {"", None} else ("", "")
                for k, v in syntax_highlighting.items()
            })

            sep = f"{self.syntax_hl['punctuation'][0]}{sep}{self.syntax_hl['punctuation'][1]}"

        self.sep = sep

        punct_map: dict[str, str | tuple[str, str]] = {"(": ("/(", "("), **{c: c for c in "'\":)[]{}"}}
        self.punct: dict[str, str] = {
            k: ((f"{self.syntax_hl['punctuation'][0]}{v[0]}{self.syntax_hl['punctuation'][1]}" if self.do_syntax_hl else v[1])
                if isinstance(v, (list, tuple)) else
                (f"{self.syntax_hl['punctuation'][0]}{v}{self.syntax_hl['punctuation'][1]}" if self.do_syntax_hl else v))
            for k, v in punct_map.items()
        }

    def __call__(self) -> str:
        return _re.sub(
            r"\s+(?=\n)", "",
            self.format_dict(self.data, 0) if isinstance(self.data, dict) else self.format_sequence(self.data, 0)
        )

    def format_value(self, value: Any, current_indent: Optional[int] = None) -> str:
        if current_indent is not None and isinstance(value, dict):
            return self.format_dict(value, current_indent + self.indent)
        elif current_indent is not None and hasattr(value, "__dict__"):
            return self.format_dict(value.__dict__, current_indent + self.indent)
        elif current_indent is not None and isinstance(value, IndexIterableTypes):
            return self.format_sequence(value, current_indent + self.indent)
        elif current_indent is not None and isinstance(value, (bytes, bytearray)):
            obj_dict = self.cls.serialize_bytes(value)
            return (
                self.format_dict(obj_dict, current_indent + self.indent) if self.as_json else (
                    f"{self.syntax_hl['type'][0]}{(k := next(iter(obj_dict)))}{self.syntax_hl['type'][1]}"
                    + self.format_sequence((obj_dict[k], obj_dict["encoding"]), current_indent + self.indent)
                    if self.do_syntax_hl else (k := next(iter(obj_dict)))
                    + self.format_sequence((obj_dict[k], obj_dict["encoding"]), current_indent + self.indent)
                )
            )
        elif isinstance(value, bool):
            val = str(value).lower() if self.as_json else str(value)
            return f"{self.syntax_hl['literal'][0]}{val}{self.syntax_hl['literal'][1]}" if self.do_syntax_hl else val
        elif isinstance(value, (int, float)):
            val = "null" if self.as_json and (_math.isinf(value) or _math.isnan(value)) else str(value)
            return f"{self.syntax_hl['number'][0]}{val}{self.syntax_hl['number'][1]}" if self.do_syntax_hl else val
        elif current_indent is not None and isinstance(value, complex):
            return (
                self.format_value(str(value).strip("()")) if self.as_json else (
                    f"{self.syntax_hl['type'][0]}complex{self.syntax_hl['type'][1]}"
                    + self.format_sequence((value.real, value.imag), current_indent + self.indent) if self.do_syntax_hl else
                    f"complex{self.format_sequence((value.real, value.imag), current_indent + self.indent)}"
                )
            )
        elif value is None:
            val = "null" if self.as_json else "None"
            return f"{self.syntax_hl['literal'][0]}{val}{self.syntax_hl['literal'][1]}" if self.do_syntax_hl else val
        else:
            return ((
                self.punct['"'] + self.syntax_hl["str"][0] + String.escape(str(value), '"') + self.syntax_hl["str"][1]
                + self.punct['"'] if self.do_syntax_hl else self.punct['"'] + String.escape(str(value), '"') + self.punct['"']
            ) if self.as_json else (
                self.punct["'"] + self.syntax_hl["str"][0] + String.escape(str(value), "'") + self.syntax_hl["str"][1]
                + self.punct["'"] if self.do_syntax_hl else self.punct["'"] + String.escape(str(value), "'") + self.punct["'"]
            ))

    def should_expand(self, seq: IndexIterable) -> bool:
        if self.compactness == 0:
            return True
        if self.compactness == 2:
            return False

        complex_types: tuple[type, ...] = (list, tuple, dict, set, frozenset)
        if self.as_json:
            complex_types += (bytes, bytearray)

        complex_items = sum(1 for item in seq if isinstance(item, complex_types))

        return complex_items > 1 \
            or (complex_items == 1 and len(seq) > 1) \
            or self.cls.chars_count(seq) + (len(seq) * len(self.sep)) > self.max_width

    def format_dict(self, d: dict, current_indent: int) -> str:
        if self.compactness == 2 or not d or not self.should_expand(list(d.values())):
            return self.punct["{"] + self.sep.join(
                f"{self.format_value(k)}{self.punct[':']} {self.format_value(v, current_indent)}" for k, v in d.items()
            ) + self.punct["}"]

        items = []
        for k, val in d.items():
            formatted_value = self.format_value(val, current_indent)
            items.append(f"{' ' * (current_indent + self.indent)}{self.format_value(k)}{self.punct[':']} {formatted_value}")

        return self.punct["{"] + "\n" + f"{self.sep}\n".join(items) + f"\n{' ' * current_indent}" + self.punct["}"]

    def format_sequence(self, seq, current_indent: int) -> str:
        if self.as_json:
            seq = list(seq)

        brackets = (self.punct["["], self.punct["]"]) if isinstance(seq, list) else (self.punct["("], self.punct[")"])

        if self.compactness == 2 or not seq or not self.should_expand(seq):
            return f"{brackets[0]}{self.sep.join(self.format_value(item, current_indent) for item in seq)}{brackets[1]}"

        items = [self.format_value(item, current_indent) for item in seq]
        formatted_items = f"{self.sep}\n".join(f'{" " * (current_indent + self.indent)}{item}' for item in items)

        return f"{brackets[0]}\n{formatted_items}\n{' ' * current_indent}{brackets[1]}"



"""
This module provides the `System` class, which includes
methods to interact with the underlying operating system.
"""

from .base.types import MissingLibsMsgs

from .format_codes import FormatCodes
from .console import Console

from typing import Optional
from mypy_extensions import mypyc_attr
import subprocess as _subprocess
import platform as _platform
import ctypes as _ctypes
import time as _time
import sys as _sys
import os as _os


@mypyc_attr(native_class=False)
class _SystemMeta(type):

    @property
    def is_elevated(cls) -> bool:
        """Whether the current process has elevated privileges or not."""
        try:
            if _os.name == "nt":
                return getattr(_ctypes, "windll").shell32.IsUserAnAdmin() != 0
            elif _os.name == "posix":
                return _os.geteuid() == 0  # type: ignore[attr-defined]
        except Exception:
            pass
        return False

    @property
    def is_win(cls) -> bool:
        """Whether the current operating system is Windows or not."""
        return _platform.system() == "Windows"


class System(metaclass=_SystemMeta):
    """This class provides methods to interact with the underlying operating system."""

    @classmethod
    def restart(cls, prompt: object = "", wait: int = 0, continue_program: bool = False, force: bool = False) -> None:
        """Restarts the system with some advanced options\n
        --------------------------------------------------------------------------------------------------
        - `prompt` -⠀the message to be displayed in the systems restart notification
        - `wait` -⠀the time to wait until restarting in seconds
        - `continue_program` -⠀whether to continue the current Python program after calling this function
        - `force` -⠀whether to force a restart even if other processes are still running"""
        if wait < 0:
            raise ValueError(f"The 'wait' parameter must be non-negative, got {wait!r}")

        _SystemRestartHelper(prompt, wait, continue_program, force)()

    @classmethod
    def check_libs(
        cls,
        lib_names: list[str],
        install_missing: bool = False,
        missing_libs_msgs: MissingLibsMsgs = {
            "found_missing": "The following required libraries are missing:",
            "should_install": "Do you want to install them now?",
        },
        confirm_install: bool = True,
    ) -> Optional[list[str]]:
        """Checks if the given list of libraries are installed and optionally installs missing libraries.\n
        ------------------------------------------------------------------------------------------------------------
        - `lib_names` -⠀a list of library names to check
        - `install_missing` -⠀whether to directly missing libraries will be installed automatically using pip
        - `missing_libs_msgs` -⠀two messages: the first one is displayed when missing libraries are found,
          the second one is the confirmation message before installing missing libraries
        - `confirm_install` -⠀whether the user will be asked for confirmation before installing missing libraries\n
        ------------------------------------------------------------------------------------------------------------
        If some libraries are missing or they could not be installed, their names will be returned as a list.
        If all libraries are installed (or were installed successfully), `None` will be returned."""
        return _SystemCheckLibsHelper(lib_names, install_missing, missing_libs_msgs, confirm_install)()

    @classmethod
    def elevate(cls, win_title: Optional[str] = None, args: Optional[list] = None) -> bool:
        """Attempts to start a new process with elevated privileges.\n
        ---------------------------------------------------------------------------------
        - `win_title` -⠀the window title of the elevated process (only on Windows)
        - `args` -⠀a list of additional arguments to be passed to the elevated process\n
        ---------------------------------------------------------------------------------
        After the elevated process started, the original process will exit.<br>
        This means, that this method has to be run at the beginning of the program or
        or else the program has to continue in a new window after elevation.\n
        ---------------------------------------------------------------------------------
        Returns `True` if the current process already has elevated privileges and raises
        a `PermissionError` if the user denied the elevation or the elevation failed."""
        if cls.is_elevated:
            return True

        args_list = args or []

        if _os.name == "nt":  # WINDOWS
            if win_title:
                args_str = f'-c "import ctypes; ctypes.windll.kernel32.SetConsoleTitleW(\\"{win_title}\\"); exec(open(\\"{_sys.argv[0]}\\").read())" {" ".join(args_list)}"'
            else:
                args_str = f'-c "exec(open(\\"{_sys.argv[0]}\\").read())" {" ".join(args_list)}'

            result = getattr(_ctypes, "windll").shell32.ShellExecuteW(None, "runas", _sys.executable, args_str, None, 1)
            if result <= 32:
                raise PermissionError("Failed to launch elevated process.")
            else:
                _sys.exit(0)

        else:  # POSIX
            cmd = ["pkexec"]
            if win_title:
                cmd.extend(["--description", win_title])
            cmd.extend([_sys.executable] + _sys.argv[1:] + args_list)

            proc = _subprocess.Popen(cmd)
            proc.wait()
            if proc.returncode != 0:
                raise PermissionError("Process elevation was denied.")
            _sys.exit(0)


class _SystemRestartHelper:
    """Internal, callable helper class to handle system restart with platform-specific logic."""

    def __init__(self, prompt: object, wait: int, continue_program: bool, force: bool):
        self.prompt = prompt
        self.wait = wait
        self.continue_program = continue_program
        self.force = force

    def __call__(self) -> None:
        if (system := _platform.system().lower()) == "windows":
            self.restart_windows()
        elif system in {"linux", "darwin"}:
            self.restart_posix()
        else:
            raise NotImplementedError(f"Restart not implemented for '{system}' systems.")

    def check_running_processes(self, command: str | list[str], skip_lines: int = 0) -> None:
        """Check if processes are running and raise error if force is False."""
        if self.force:
            return

        if isinstance(command, str):
            output = _subprocess.check_output(command, shell=True).decode()
        else:
            output = _subprocess.check_output(command).decode()

        processes = [line for line in output.splitlines()[skip_lines:] if line.strip()]
        if len(processes) > 2:  # EXCLUDING PYTHON AND SHELL PROCESSES
            raise RuntimeError("Processes are still running.\nTo restart anyway set parameter 'force' to True.")

    def restart_windows(self) -> None:
        """Handle Windows system restart."""
        self.check_running_processes("tasklist", skip_lines=3)

        if self.prompt:
            _os.system(f'shutdown /r /t {self.wait} /c "{self.prompt}"')
        else:
            _os.system("shutdown /r /t 0")

        if self.continue_program:
            self.wait_for_restart()

    def restart_posix(self) -> None:
        """Handle Linux/macOS system restart."""
        self.check_running_processes(["ps", "-A"], skip_lines=1)

        if self.prompt:
            _subprocess.Popen(["notify-send", "System Restart", str(self.prompt)])
            _time.sleep(self.wait)

        try:
            _subprocess.run(["sudo", "shutdown", "-r", "now"])
        except _subprocess.CalledProcessError:
            raise PermissionError("Failed to restart: insufficient privileges.\nEnsure sudo permissions are granted.")

        if self.continue_program:
            self.wait_for_restart()

    def wait_for_restart(self) -> None:
        """Wait and print message before restart."""
        print(f"Restarting in {self.wait} seconds...")
        _time.sleep(self.wait)


class _SystemCheckLibsHelper:
    """Internal, callable helper class to check and install missing Python libraries."""

    def __init__(
        self,
        lib_names: list[str],
        install_missing: bool,
        missing_libs_msgs: MissingLibsMsgs,
        confirm_install: bool,
    ):
        self.lib_names = lib_names
        self.install_missing = install_missing
        self.missing_libs_msgs = missing_libs_msgs
        self.confirm_install = confirm_install

    def __call__(self) -> Optional[list[str]]:
        missing = self.find_missing_libs()

        if not missing:
            return None
        elif not self.install_missing:
            return missing

        if self.confirm_install and not self.confirm_installation(missing):
            return missing

        return self.install_libs(missing)

    def find_missing_libs(self) -> list[str]:
        """Find which libraries are missing."""
        missing = []
        for lib in self.lib_names:
            try:
                __import__(lib)
            except ImportError:
                missing.append(lib)
        return missing

    def confirm_installation(self, missing: list[str]) -> bool:
        """Ask user for confirmation before installing libraries."""
        FormatCodes.print(f"[b]({self.missing_libs_msgs['found_missing']})")
        for lib in missing:
            FormatCodes.print(f" [dim](•) [i]{lib}[_i]")
        print()
        return Console.confirm(self.missing_libs_msgs["should_install"], end="\n")

    def install_libs(self, missing: list[str]) -> Optional[list[str]]:
        """Install missing libraries using pip."""
        for lib in missing[:]:
            try:
                _subprocess.check_call([_sys.executable, "-m", "pip", "install", lib])
                missing.remove(lib)
            except _subprocess.CalledProcessError:
                pass

        return None if len(missing) == 0 else missing



"""
This module provides the `Regex` class, which includes methods
to dynamically generate complex regex patterns for common use cases.
"""

from typing import Optional
from mypy_extensions import mypyc_attr
import regex as _rx
import re as _re


class Regex:
    """This class provides methods to dynamically generate complex regex patterns for common use cases."""

    @classmethod
    def quotes(cls) -> str:
        """Matches pairs of quotes. (strings)\n
        --------------------------------------------------------------------------------
        Will create two named groups:
        - `quote` the quote type (single or double)
        - `string` everything inside the found quote pair\n
        ---------------------------------------------------------------------------------
        Attention: Requires non-standard library `regex`, not standard library `re`!"""
        return r"""(?P<quote>["'])(?P<string>(?:\\.|(?!\g<quote>).)*?)\g<quote>"""

    @classmethod
    def brackets(
        cls,
        bracket1: str = "(",
        bracket2: str = ")",
        is_group: bool = False,
        strip_spaces: bool = False,
        ignore_in_strings: bool = True,
    ) -> str:
        """Matches everything inside pairs of brackets, including other nested brackets.\n
        ---------------------------------------------------------------------------------------
        - `bracket1` -⠀the opening bracket (e.g. `(`, `{`, `[`, …)
        - `bracket2` -⠀the closing bracket (e.g. `)`, `}`, `]`, …)
        - `is_group` -⠀whether to create a capturing group for the content inside the brackets
        - `strip_spaces` -⠀whether to strip spaces from the bracket content or not
        - `ignore_in_strings` -⠀whether to ignore closing brackets that are inside
          strings/quotes (e.g. `'…)…'` or `"…)…"`)\n
        ---------------------------------------------------------------------------------------
        Attention: Requires non-standard library `regex`, not standard library `re`!"""
        g = "" if is_group else "?:"
        b1 = _rx.escape(bracket1) if len(bracket1) == 1 else bracket1
        b2 = _rx.escape(bracket2) if len(bracket2) == 1 else bracket2
        s1 = r"\s*" if strip_spaces else ""
        s2 = "" if strip_spaces else r"\s*"

        if ignore_in_strings:
            return cls._clean( \
                rf"""{b1}{s1}({g}{s2}(?:
                    [^{b1}{b2}"']
                    |"(?:\\.|[^"\\])*"
                    |'(?:\\.|[^'\\])*'
                    |{b1}(?:
                        [^{b1}{b2}"']
                        |"(?:\\.|[^"\\])*"
                        |'(?:\\.|[^'\\])*'
                        |(?R)
                    )*{b2}
                )*{s2}){s1}{b2}"""
            )
        else:
            return cls._clean( \
                rf"""{b1}{s1}({g}{s2}(?:
                    [^{b1}{b2}]
                    |{b1}(?:
                        [^{b1}{b2}]
                        |(?R)
                    )*{b2}
                )*{s2}){s1}{b2}"""
            )

    @classmethod
    def outside_strings(cls, pattern: str = r".*") -> str:
        """Matches the `pattern` only when it is not found inside a string (`'…'` or `"…"`)."""
        return rf"""(?<!["'])(?:{pattern})(?!["'])"""

    @classmethod
    def all_except(cls, disallowed_pattern: str, ignore_pattern: str = "", is_group: bool = False) -> str:
        """Matches everything up to the `disallowed_pattern`, unless the
        `disallowed_pattern` is found inside a string/quotes (`'…'` or `"…"`).\n
        -------------------------------------------------------------------------------------
        - `disallowed_pattern` -⠀the pattern that is not allowed to be matched
        - `ignore_pattern` -⠀a pattern that, if found, will make the regex ignore the
          `disallowed_pattern` (even if it contains the `disallowed_pattern` inside it):<br>
          For example if `disallowed_pattern` is `>` and `ignore_pattern` is `->`,
          the `->`-arrows will be allowed, even though they have `>` in them.
        - `is_group` -⠀whether to create a capturing group for the matched content"""
        g = "" if is_group else "?:"

        return cls._clean( \
            rf"""({g}
                (?:(?!{ignore_pattern}).)*
                (?:(?!{cls.outside_strings(disallowed_pattern)}).)*
            )"""
        )

    @classmethod
    def func_call(cls, func_name: Optional[str] = None) -> str:
        """Match a function call, and get back two groups:
        1. function name
        2. the function's arguments\n
        If no `func_name` is given, it will match any function call.\n
        ---------------------------------------------------------------------------------
        Attention: Requires non-standard library `regex`, not standard library `re`!"""
        if func_name in {"", None}:
            func_name = r"[\w_]+"

        return rf"""(?<=\b)({func_name})\s*{cls.brackets("(", ")", is_group=True)}"""

    @classmethod
    def rgba_str(cls, fix_sep: Optional[str] = ",", allow_alpha: bool = True) -> str:
        """Matches an RGBA color inside a string.\n
        ----------------------------------------------------------------------------------
        - `fix_sep` -⠀the fixed separator between the RGBA values (e.g. `,`, `;` …)<br>
          If set to nothing or `None`, any char that is not a letter or number
          can be used to separate the RGBA values, including just a space.
        - `allow_alpha` -⠀whether to include the alpha channel in the match\n
        ----------------------------------------------------------------------------------
        The RGBA color can be in the formats (for `fix_sep = ','`):
        - `rgba(r, g, b)`
        - `rgba(r, g, b, a)` (if `allow_alpha=True`)
        - `(r, g, b)`
        - `(r, g, b, a)` (if `allow_alpha=True`)
        - `r, g, b`
        - `r, g, b, a` (if `allow_alpha=True`)\n
        #### Valid ranges:
        - `r` 0-255 (int: red)
        - `g` 0-255 (int: green)
        - `b` 0-255 (int: blue)
        - `a` 0.0-1.0 (float: opacity)"""
        fix_sep = _re.escape(fix_sep) if isinstance(fix_sep, str) else r"[^0-9A-Z]"

        rgb_part = rf"""((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))
            (?:\s*{fix_sep}\s*)((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))
            (?:\s*{fix_sep}\s*)((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))"""

        if allow_alpha:
            return cls._clean( \
                rf"""(?ix)(?:rgb|rgba)?\s*(?:
                    \(?\s*{rgb_part}
                        (?:(?:\s*{fix_sep}\s*)((?:0*(?:0?\.[0-9]+|1\.0+|[0-9]+\.[0-9]+|[0-9]+))))?
                    \s*\)?
                )"""
            )
        else:
            return cls._clean( \
                rf"""(?ix)(?:rgb|rgba)?\s*(?:
                    \(?\s*{rgb_part}\s*\)?
                )"""
            )

    @classmethod
    def hsla_str(cls, fix_sep: Optional[str] = ",", allow_alpha: bool = True) -> str:
        """Matches a HSLA color inside a string.\n
        ----------------------------------------------------------------------------------
        - `fix_sep` -⠀the fixed separator between the HSLA values (e.g. `,`, `;` …)<br>
          If set to nothing or `None`, any char that is not a letter or number
          can be used to separate the HSLA values, including just a space.
        - `allow_alpha` -⠀whether to include the alpha channel in the match\n
        ----------------------------------------------------------------------------------
        The HSLA color can be in the formats (for `fix_sep = ','`):
        - `hsla(h, s, l)`
        - `hsla(h, s, l, a)` (if `allow_alpha=True`)
        - `(h, s, l)`
        - `(h, s, l, a)` (if `allow_alpha=True`)
        - `h, s, l`
        - `h, s, l, a` (if `allow_alpha=True`)\n
        #### Valid ranges:
        - `h` 0-360 (int: hue)
        - `s` 0-100 (int: saturation)
        - `l` 0-100 (int: lightness)
        - `a` 0.0-1.0 (float: opacity)"""
        fix_sep = _re.escape(fix_sep) if isinstance(fix_sep, str) else r"[^0-9A-Z]"

        hsl_part = rf"""((?:0*(?:360|3[0-5][0-9]|[12][0-9][0-9]|[1-9]?[0-9])))(?:\s*°)?
            (?:\s*{fix_sep}\s*)((?:0*(?:100|[1-9][0-9]|[0-9])))(?:\s*%)?
            (?:\s*{fix_sep}\s*)((?:0*(?:100|[1-9][0-9]|[0-9])))(?:\s*%)?"""

        if allow_alpha:
            return cls._clean( \
                rf"""(?ix)(?:hsl|hsla)?\s*(?:
                    \(?\s*{hsl_part}
                        (?:(?:\s*{fix_sep}\s*)((?:0*(?:0?\.[0-9]+|1\.0+|[0-9]+\.[0-9]+|[0-9]+))))?
                    \s*\)?
                )"""
            )
        else:
            return cls._clean( \
                rf"""(?ix)(?:hsl|hsla)?\s*(?:
                    \(?\s*{hsl_part}\s*\)?
                )"""
            )

    @classmethod
    def hexa_str(cls, allow_alpha: bool = True) -> str:
        """Matches a HEXA color inside a string.\n
        ----------------------------------------------------------------------
        - `allow_alpha` -⠀whether to include the alpha channel in the match\n
        ----------------------------------------------------------------------
        The HEXA color can be in the formats (prefix `#`, `0x` or no prefix):
        - `RGB`
        - `RGBA` (if `allow_alpha=True`)
        - `RRGGBB`
        - `RRGGBBAA` (if `allow_alpha=True`)\n
        #### Valid ranges:
        every channel from 0-9 and A-F (case insensitive)"""
        return r"(?i)(?:#|0x)?([0-9A-F]{8}|[0-9A-F]{6}|[0-9A-F]{4}|[0-9A-F]{3})" \
            if allow_alpha else r"(?i)(?:#|0x)?([0-9A-F]{6}|[0-9A-F]{3})"

    @classmethod
    def _clean(cls, pattern: str) -> str:
        """Internal method to make a multiline-string regex pattern into a single-line pattern."""
        return "".join(line.strip() for line in pattern.splitlines()).strip()


@mypyc_attr(native_class=False)
class LazyRegex:
    """A class that lazily compiles and caches regex patterns on first access.\n
    --------------------------------------------------------------------------------
    - `**patterns` -⠀keyword arguments where the key is the name of the pattern and
      the value is the regex pattern string to compile\n
    --------------------------------------------------------------------------------
    #### Example usage:
    ```python
    PATTERNS = LazyRegex(
        email=r"(?i)[A-Z0-9._%+-]+@[A-Z0-9.-]+\\.[A-Z]{2,}",
        phone=r"\\+?\\d{1,3}[-.\\s]?\\(?\\d{1,4}\\)?[-.\\s]?\\d{1,4}[-.\\s]?\\d{1,9}",
    )

    email_pattern = PATTERNS.email  # Compiles and caches the EMAIL pattern
    phone_pattern = PATTERNS.phone  # Compiles and caches the PHONE pattern
    ```"""

    def __init__(self, **patterns: str):
        self._patterns = patterns

    def __getattr__(self, name: str) -> _rx.Pattern:
        if name in self._patterns:
            setattr(self, name, compiled := _rx.compile(self._patterns[name]))
            return compiled

        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")



from xulbux.data import Data

import pytest

# ! DON'T CHANGE THIS DATA !
d_comments = {
    "key1": [
        ">> COMMENT IN THE BEGINNING OF THE STRING <<  value1",
        "value2  >> COMMENT IN THE END OF THE STRING",
        "val>> COMMENT IN THE MIDDLE OF THE STRING <<ue3",
        ">> FULL VALUE IS A COMMENT  value4",
    ],
    ">> FULL KEY + ALL ITS VALUES ARE A COMMENT  key2": ["value", "value", "value"],
    "key3":
    ">> ALL THE KEYS VALUES ARE COMMENTS  value",
}

d1_equal = {
    "key1": ["value1", "value2", "value3", ["value1", "value2", "value3"]],
    "key2": ["value1", "value2", "value3", ["value1", "value2", "value3"]],
    "key3": "value",
}
d2_equal = {
    "key1": ["value1", "value2", "value3", ["value1", "value2", "value3"]],
    "key2": ["value1", "value2", "value3", ["value1", "value2", "value3"]],
    "key3": "CHANGED value",
}

d1_path_id = {"healthy": {"fruit": ["apples", "bananas", "oranges"], "vegetables": ["carrots", "broccoli", "celery"]}}
d2_path_id = {"school": {"material": ["pencil", "paper", "rubber"], "subjects": ["math", "science", "history"]}}

#
################################################## Data TESTS ##################################################


def test_serialize_bytes():
    utf8_bytes = b"Hello"
    utf8_bytearray = bytearray(b"World")
    non_utf8_bytes = b"\x80abc"

    assert Data.serialize_bytes(utf8_bytes) == {"bytes": "Hello", "encoding": "utf-8"}
    assert Data.serialize_bytes(utf8_bytearray) == {"bytearray": "World", "encoding": "utf-8"}
    serialized_non_utf8 = Data.serialize_bytes(non_utf8_bytes)
    assert serialized_non_utf8["encoding"] == "base64"
    import base64
    assert base64.b64decode(serialized_non_utf8["bytes"]).decode("latin-1") == non_utf8_bytes.decode("latin-1")


def test_deserialize_bytes():
    utf8_serialized_bytes = {"bytes": "Hello", "encoding": "utf-8"}
    utf8_serialized_bytearray = {"bytearray": "World", "encoding": "utf-8"}
    import base64
    non_utf8_bytes = b"\x80abc"
    base64_encoded = base64.b64encode(non_utf8_bytes).decode("utf-8")
    base64_serialized_bytes = {"bytes": base64_encoded, "encoding": "base64"}

    assert Data.deserialize_bytes(utf8_serialized_bytes) == b"Hello"
    assert Data.deserialize_bytes(utf8_serialized_bytearray) == bytearray(b"World")
    assert Data.deserialize_bytes(base64_serialized_bytes) == non_utf8_bytes

    with pytest.raises(ValueError):
        Data.deserialize_bytes({"bytes": "abc", "encoding": "unknown"})
    with pytest.raises(ValueError):
        Data.deserialize_bytes({"wrong_key": "abc", "encoding": "utf-8"})


@pytest.mark.parametrize(
    "input_data, expected_count", [
        (["a", "bc", "def"], 6),
        (("a", "bc", "def"), 6),
        ({"a", "bc", "def"}, 6),
        ({"k1": "v1", "k2": "v2"}, 8),
        (["ab", ["c", "d"]], 4),
        ({"k": ["v1", "v2"]}, 5),
        ([], 0),
        ({}, 0),
    ]
)
def test_chars_count(input_data, expected_count):
    assert Data.chars_count(input_data) == expected_count


@pytest.mark.parametrize(
    "input_data, expected_output", [
        (["  a  ", " b ", "c"], ["a", "b", "c"]),
        (("  a  ", " b ", "c"), ("a", "b", "c")),
        ({"  a  ": "  v1 ", " b ": "v2"}, {"a": "v1", "b": "v2"}),
        ([" a ", [" b ", " c"]], ["a", ["b", "c"]]),
    ]
)
def test_strip(input_data, expected_output):
    assert Data.strip(input_data) == expected_output


@pytest.mark.parametrize(
    "input_data, spaces_are_empty, expected_output", [
        (["a", "", "b", None, "  "], False, ["a", "b", "  "]),
        (["a", "", "b", None, "  "], True, ["a", "b"]),
        (("a", "", "b", None, "  "), False, ("a", "b", "  ")),
        (("a", "", "b", None, "  "), True, ("a", "b")),
        ({"k1": "a", "k2": "", "k3": "b", "k4": None, "k5": "  "}, False, {"k1": "a", "k3": "b", "k5": "  "}),
        ({"k1": "a", "k2": "", "k3": "b", "k4": None, "k5": "  "}, True, {"k1": "a", "k3": "b"}),
        (["a", ["", "b"], "c"], False, ["a", ["b"], "c"]),
        (["a", ["", "b"], "c"], True, ["a", ["b"], "c"]),
        (["a", {"x": "", "y": "b"}, "c"], False, ["a", {"y": "b"}, "c"]),
        (["a", {"x": "", "y": "b"}, "c"], True, ["a", {"y": "b"}, "c"]),
        (["a", [], {}], False, ["a"]),
    ]
)
def test_remove_empty_items(input_data, spaces_are_empty, expected_output):
    assert Data.remove_empty_items(input_data, spaces_are_empty) == expected_output


@pytest.mark.parametrize(
    "input_data, expected_output", [
        (["a", "b", "a", "c", "b"], ["a", "b", "c"]),
        (("a", "b", "a", "c", "b"), ("a", "b", "c")),
        ({"a", "b", "a", "c", "b"}, {"a", "b", "c"}),
        ({"k1": "a", "k2": "b", "k3": "a"}, {"k1": "a", "k2": "b", "k3": "a"}),
        (["a", ["b", "b"], "c"], ["a", ["b"], "c"]),
        ({"k": ["v", "v"]}, {"k": ["v"]}),
    ]
)
def test_remove_duplicates(input_data, expected_output):
    assert Data.remove_duplicates(input_data) == expected_output


def test_remove_comments():
    assert Data.remove_comments(
        d_comments, comment_sep="__"
    ) == {
        "key1": ["value1", "value2", "val__ue3"],
        "key3": None,
    }


def test_is_equal():
    assert Data.is_equal(d1_equal, d2_equal) is False
    assert Data.is_equal(d1_equal, d2_equal, ignore_paths="key3") is True


def test_get_path_id():
    id1, id2 = Data.get_path_id(
        d1_path_id, ["healthy->fruit->bananas", "healthy->vegetables->2"]
    )  # type: ignore[return-value]
    assert id1 == "1>001"
    assert id2 == "1>012"
    assert Data.get_value_by_path_id(d1_path_id, id1) == "bananas"
    assert Data.get_value_by_path_id(d1_path_id, id2) == "celery"
    assert Data.set_value_by_path_id(d2_path_id, {id1: "NEW1", id2: "NEW2"}) == {
        "school": {"material": ["pencil", "NEW1", "rubber"], "subjects": ["math", "science", "NEW2"]}
    }


def test_get_value_by_path_id():
    data = {"a": [1, {"b": "c"}], "d": ("e", "f")}
    path_id_1 = str(Data.get_path_id(data, "a->1->b"))
    path_id_2 = str(Data.get_path_id(data, "d->1"))

    assert path_id_1 == "1>010"
    assert path_id_2 == "1>11"
    assert Data.get_value_by_path_id(data, path_id_1) == "c"
    assert Data.get_value_by_path_id(data, path_id_2) == "f"
    assert Data.get_value_by_path_id(data, path_id_1, get_key=True) == "b"
    assert Data.get_value_by_path_id(data, path_id_2, get_key=True) == "d"

    with pytest.raises(ValueError):
        Data.get_value_by_path_id(data, "invalid_id")
    with pytest.raises(IndexError):
        Data.get_value_by_path_id({"a": [1]}, "1>01")


def test_set_value_by_path_id():
    data = {"a": [1, {"b": "c"}], "d": ("e", "f")}
    path_id_c = Data.get_path_id(data, "a->1->b")
    path_id_f = Data.get_path_id(data, "d->1")

    updated_data = Data.set_value_by_path_id(data, {path_id_c: "NEW_C", path_id_f: "NEW_F"})  # type: ignore[assignment]
    expected_data = {"a": [1, {"b": "NEW_C"}], "d": ("e", "NEW_F")}
    assert updated_data == expected_data

    updated_data_types = Data.set_value_by_path_id(data, {path_id_c: [1, 2], path_id_f: {"x": 1}})  # type: ignore[assignment]
    expected_data_types = {"a": [1, {"b": [1, 2]}], "d": ("e", {"x": 1})}
    assert updated_data_types == expected_data_types

    with pytest.raises(ValueError):
        Data.set_value_by_path_id(data, {"invalid": "value"})

    with pytest.raises(ValueError):
        Data.set_value_by_path_id(data, {})


@pytest.mark.parametrize(
    "data, indent, compactness, max_width, sep, as_json, expected_str", [
        ([1, 2, 3], 4, 1, 80, ", ", False, "[1, 2, 3]"),
        ({"a": 1, "b": 2}, 4, 1, 80, ", ", False, "{'a': 1, 'b': 2}"),
        ({"a": [1, 2], "b": {"c": 3}}, 4, 1, 80, ", ", False, "{\n    'a': [1, 2],\n    'b': {'c': 3}\n}"),
        ({"a": [1, 2], "b": {"c": 3}
          }, 4, 0, 80, ", ", False, "{\n    'a': [\n        1,\n        2\n    ],\n    'b': {\n        'c': 3\n    }\n}"),
        ({"a": [1, 2], "b": {"c": 3}}, 4, 2, 80, ", ", False, "{'a': [1, 2], 'b': {'c': 3}}"),
        ([1, [2, 3]], 2, 1, 80, ", ", False, "[\n  1,\n  [2, 3]\n]"),
        ({"ultralongkeyname": [1, None, False]}, 4, 1, 20, ", ", False, "{'ultralongkeyname': [1, None, False]}"),
        ([1, 2, 3], 4, 1, 80, "; ", False, "[1; 2; 3]"),
        ({"a": True, "b": None, "c": [1, 2.5]}, 4, 2, 80, ", ", True, '{"a": true, "b": null, "c": [1, 2.5]}'),
        ({"data": b"hello"}, 2, 0, 80, ", ", True, '{\n  "data": {\n    "bytes": "hello",\n    "encoding": "utf-8"\n  }\n}'),
        ({"data": b"hello"}, 4, 1, 80, ", ", False, "{'data': bytes('hello', 'utf-8')}"),
    ]
)
def test_render(data, indent, compactness, max_width, sep, as_json, expected_str):
    result = Data.render(data, indent, compactness, max_width, sep, as_json, syntax_highlighting=False)
    normalized_result = "\n".join(line.rstrip() for line in result.splitlines())
    normalized_expected = "\n".join(line.rstrip() for line in expected_str.splitlines())
    assert normalized_result == normalized_expected



from xulbux.base.consts import ANSI
from xulbux.format_codes import FormatCodes


black = ANSI.SEQ_COLOR.format(0, 0, 0)
bg_red = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP['bg:red']}{ANSI.END}"
default = ANSI.SEQ_COLOR.format(255, 255, 255)
orange = ANSI.SEQ_COLOR.format(255, 136, 119)

bold = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('bold', 'b')]}{ANSI.END}"
invert = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('inverse', 'invert', 'in')]}{ANSI.END}"
italic = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('italic', 'i')]}{ANSI.END}"
underline = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('underline', 'u')]}{ANSI.END}"

reset = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP['_']}{ANSI.END}"
reset_bg = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('_background', '_bg')]}{ANSI.END}"
reset_bold = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('_bold', '_b')]}{ANSI.END}"
reset_color = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('_color', '_c')]}{ANSI.END}"
reset_italic = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('_italic', '_i')]}{ANSI.END}"
reset_invert = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('_inverse', '_invert', '_in')]}{ANSI.END}"
reset_underline = f"{ANSI.CHAR}{ANSI.START}{ANSI.CODES_MAP[('_underline', '_u')]}{ANSI.END}"

#
################################################## FormatCodes TESTS ##################################################


def test_to_ansi():
    assert (
        FormatCodes.to_ansi("[b|#000|bg:red](He[in](l)lo) [[i|u|#F87](world)][default]![_]",
                            default_color="#FFF") == f"{default}{bold}{black}{bg_red}" + "He" + invert + "l" + reset_invert
        + "lo" + f"{reset_bold}{default}{reset_bg}" + " [" + f"{italic}{underline}{orange}" + "world"
        + f"{reset_italic}{reset_underline}{default}" + "]" + default + "!" + reset
    )


def test_escape_ansi():
    ansi_string = f"{bold}Hello {orange}World!{reset}"
    escaped_string = ansi_string.replace(ANSI.CHAR, ANSI.CHAR_ESCAPED)
    assert FormatCodes.escape_ansi(ansi_string) == escaped_string


def test_escape():
    # TEST BASIC FORMATTING CODES
    assert FormatCodes.escape("[b]Hello[_]") == "[/b]Hello[/_]"
    assert FormatCodes.escape("[bold|italic]Text[_]") == "[/bold|italic]Text[/_]"

    # TEST WITH COLORS
    assert FormatCodes.escape("[#F87]Hello[_]") == "[/#F87]Hello[/_]"
    assert FormatCodes.escape("[rgb(255, 136, 119)]Hello[_]") == "[/rgb(255, 136, 119)]Hello[/_]"

    # TEST WITH DEFAULT COLOR
    assert FormatCodes.escape("[default]Hello", default_color="#FFF") == "[/default]Hello"
    assert FormatCodes.escape("[bg:default]Hello", default_color="#FFF") == "[/bg:default]Hello"

    # TEST WITH * FORMATTING CODE
    assert FormatCodes.escape("[*]Hello", default_color="#FFF") == "[/*]Hello"
    assert FormatCodes.escape("[b|*]Hello", default_color="#FFF") == "[/b|*]Hello"

    # TEST WITH AUTO-RESET
    assert FormatCodes.escape("[b](Hello)") == "[/b](Hello)"
    assert FormatCodes.escape("[*](Hello)", default_color="#FFF") == "[/*](Hello)"

    # TEST INVALID FORMATTING CODES (SHOULD REMAIN UNCHANGED)
    assert FormatCodes.escape("[invalid]Hello") == "[invalid]Hello"
    assert FormatCodes.escape("[default]Hello") == "[default]Hello"  # NO 'default_color'
    assert FormatCodes.escape("[*]Hello") == "[/*]Hello"  # NO 'default_color'

    # TEST ALREADY ESCAPED CODES
    assert FormatCodes.escape("[/b]Hello") == "[/b]Hello"
    assert FormatCodes.escape("[/*]Hello", default_color="#FFF") == "[/*]Hello"

    # TEST WITH BRIGHTNESS MODIFIERS
    assert FormatCodes.escape("[l]Hello", default_color="#FFF") == "[/l]Hello"
    assert FormatCodes.escape("[ll]Hello", default_color="#FFF") == "[/ll]Hello"
    assert FormatCodes.escape("[+]Hello", default_color="#FFF") == "[/+]Hello"
    assert FormatCodes.escape("[++]Hello", default_color="#FFF") == "[/++]Hello"
    assert FormatCodes.escape("[d]Hello", default_color="#FFF") == "[/d]Hello"
    assert FormatCodes.escape("[dd]Hello", default_color="#FFF") == "[/dd]Hello"
    assert FormatCodes.escape("[-]Hello", default_color="#FFF") == "[/-]Hello"
    assert FormatCodes.escape("[--]Hello", default_color="#FFF") == "[/--]Hello"


def test_remove_ansi():
    ansi_string = f"{bold}Hello {orange}World!{reset}"
    clean_string = "Hello World!"
    assert FormatCodes.remove_ansi(ansi_string) == clean_string


def test_remove_ansi_with_removals():
    ansi_string = f"{bold}Hello\n{orange}World!{reset}"
    clean_string = "Hello\nWorld!"
    removals = ((0, bold), (6, orange), (12, reset))
    assert FormatCodes.remove_ansi(ansi_string, get_removals=True) == (clean_string, removals)
    removals = ((0, bold), (5, orange), (11, reset))
    assert FormatCodes.remove_ansi(ansi_string, get_removals=True, _ignore_linebreaks=True) == (clean_string, removals)


def test_remove_formatting():
    format_string = "[b](Hello [#F87](World!))"
    clean_string = "Hello World!"
    assert FormatCodes.remove(format_string) == clean_string


def test_remove_formatting_with_removals():
    format_string = "[b](Hello [#F87](World!))"
    clean_string = "Hello World!"
    removals = ((0, default), (0, bold), (6, orange), (12, default), (12, reset_bold))
    assert FormatCodes.remove(format_string, default_color="#FFF", get_removals=True) == (clean_string, removals)
    format_string = "[b](Hello)\n[#F87](World!)"
    clean_string = "Hello\nWorld!"
    removals = ((0, default), (0, bold), (5, reset_bold), (6, orange), (12, default))
    assert FormatCodes.remove(format_string, default_color="#FFF", get_removals=True) == (clean_string, removals)
    removals = ((0, default), (0, bold), (5, reset_bold), (5, orange), (11, default))
    assert FormatCodes.remove(
        format_string, default_color="#FFF", get_removals=True, _ignore_linebreaks=True
    ) == (clean_string, removals)



"""
This module provides the `Code` class, which offers methods to work with code strings.
"""

from .string import String
from .regex import Regex
from .data import Data

import regex as _rx


class Code:
    """This class includes methods to work with code strings."""

    @classmethod
    def add_indent(cls, code: str, indent: int) -> str:
        """Adds `indent` spaces at the beginning of each line.\n
        --------------------------------------------------------------------------
        - `code` -⠀the code to indent
        - `indent` -⠀the amount of spaces to add at the beginning of each line"""
        if indent < 0:
            raise ValueError(f"The 'indent' parameter must be non-negative, got {indent!r}")

        return "\n".join(" " * indent + line for line in code.splitlines())

    @classmethod
    def get_tab_spaces(cls, code: str) -> int:
        """Will try to get the amount of spaces used for indentation.\n
        ----------------------------------------------------------------
        - `code` -⠀the code to analyze"""
        indents = [len(line) - len(line.lstrip()) for line in String.get_lines(code, remove_empty_lines=True)]
        return min(non_zero_indents) if (non_zero_indents := [i for i in indents if i > 0]) else 0

    @classmethod
    def change_tab_size(cls, code: str, new_tab_size: int, remove_empty_lines: bool = False) -> str:
        """Replaces all tabs with `new_tab_size` spaces.\n
        --------------------------------------------------------------------------------
        - `code` -⠀the code to modify the tab size of
        - `new_tab_size` -⠀the new amount of spaces per tab
        - `remove_empty_lines` -⠀is true, empty lines will be removed in the process"""
        if new_tab_size < 0:
            raise ValueError(f"The 'new_tab_size' parameter must be non-negative, got {new_tab_size!r}")

        code_lines = String.get_lines(code, remove_empty_lines=remove_empty_lines)

        if ((tab_spaces := cls.get_tab_spaces(code)) == new_tab_size) or tab_spaces == 0:
            if remove_empty_lines:
                return "\n".join(code_lines)
            return code

        result = []
        for line in code_lines:
            indent_level = (len(line) - len(stripped := line.lstrip())) // tab_spaces
            result.append((" " * (indent_level * new_tab_size)) + stripped)

        return "\n".join(result)

    @classmethod
    def get_func_calls(cls, code: str) -> list:
        """Will try to get all function calls and return them as a list.\n
        -------------------------------------------------------------------
        - `code` -⠀the code to analyze"""
        nested_func_calls = []

        for _, func_attrs in (funcs := _rx.findall(r"(?i)" + Regex.func_call(), code)):
            if (nested_calls := _rx.findall(r"(?i)" + Regex.func_call(), func_attrs)):
                nested_func_calls.extend(nested_calls)

        return list(Data.remove_duplicates(funcs + nested_func_calls))

    @classmethod
    def is_js(cls, code: str, funcs: set[str] = {"__", "$t", "$lang"}) -> bool:
        """Will check if the code is very likely to be JavaScript.\n
        -------------------------------------------------------------
        - `code` -⠀the code to analyze
        - `funcs` -⠀a list of custom function names to check for"""
        if len(code.strip()) < 3:
            return False

        for func in funcs:
            if _rx.match(r"^[\s\n]*" + _rx.escape(func) + r"\([^\)]*\)[\s\n]*$", code):
                return True

        direct_js_patterns = [
            r"""^[\s\n]*\$\(["'][^"']+["']\)\.[\w]+\([^\)]*\);?[\s\n]*$""",  # jQuery calls
            r"^[\s\n]*\$\.[a-zA-Z]\w*\([^\)]*\);?[\s\n]*$",  # $.ajax(), etc.
            r"^[\s\n]*\(\s*function\s*\(\)\s*\{.*\}\s*\)\(\);?[\s\n]*$",  # IIFE
            r"^[\s\n]*document\.[a-zA-Z]\w*\([^\)]*\);?[\s\n]*$",  # document.getElementById()
            r"^[\s\n]*window\.[a-zA-Z]\w*\([^\)]*\);?[\s\n]*$",  # window.alert()
            r"^[\s\n]*console\.[a-zA-Z]\w*\([^\)]*\);?[\s\n]*$",  # console.log()
        ]
        for pattern in direct_js_patterns:
            if _rx.match(pattern, code):
                return True

        arrow_function_patterns = [
            r"^[\s\n]*\b[\w_]+\s*=\s*\([^\)]*\)\s*=>\s*[^;{]*[;]?[\s\n]*$",  # const x = (y) => y*2;
            r"^[\s\n]*\b[\w_]+\s*=\s*[\w_]+\s*=>\s*[^;{]*[;]?[\s\n]*$",  # const x = y => y*2;
            r"^[\s\n]*\(\s*[\w_,\s]+\s*\)\s*=>\s*[^;{]*[;]?[\s\n]*$",  # (x) => x*2
            r"^[\s\n]*[\w_]+\s*=>\s*[^;{]*[;]?[\s\n]*$",  # x => x*2
        ]
        for pattern in arrow_function_patterns:
            if _rx.match(pattern, code):
                return True

        js_score = 0.0
        funcs_pattern = r"(" + "|".join(_rx.escape(f) for f in funcs) + r")" + Regex.brackets("()")
        js_indicators: list[tuple[str, float]] = [
            (r"\b(var|let|const)\s+[\w_$]+", 2.0),  # JS VARIABLE DECLARATIONS
            (r"\$[\w_$]+\s*=", 2.0),  # jQuery-STYLE VARIABLES
            (r"\$[\w_$]+\s*\(", 2.0),  # jQuery FUNCTION CALLS
            (r"\bfunction\s*[\w_$]*\s*\(", 2.0),  # FUNCTION DECLARATIONS
            (r"[\w_$]+\s*=\s*function\s*\(", 2.0),  # FUNCTION ASSIGNMENTS
            (r"\b[\w_$]+\s*=>\s*[\{\(]", 2.0),  # ARROW FUNCTIONS
            (r"\(function\s*\(\)\s*\{", 2.0),  # IIFE PATTERN
            (funcs_pattern, 2.0),  # CUSTOM PREDEFINED FUNCTIONS
            (r"\b(true|false|null|undefined)\b", 1.0),  # JS LITERALS
            (r"===|!==|\+\+|--|\|\||&&", 1.5),  # JS-SPECIFIC OPERATORS
            (r"\bnew\s+[\w_$]+\s*\(", 1.5),  # OBJECT INSTANTIATION WITH NEW
            (r"\b(document|window|console|Math|Array|Object|String|Number)\.", 2.0),  # JS OBJECTS
            (r"\basync\s+function|\bawait\b", 2.0),  # ASYNC/AWAIT
            (r"\b(if|for|while|switch)\s*\([^)]*\)\s*\{", 1.0),  # CONTROL STRUCTURES WITH BRACES
            (r"\btry\s*\{[^}]*\}\s*catch\s*\(", 1.5),  # TRY-CATCH
            (r";[\s\n]*$", 0.5),  # SEMICOLON LINE ENDINGS
        ]

        line_endings = [line.strip() for line in code.splitlines() if line.strip()]
        if (semicolon_endings := sum(1 for line in line_endings if line.endswith(";"))) >= 1:
            js_score += min(semicolon_endings, 2)
        if (opening_braces := code.count("{")) > 0 and opening_braces == code.count("}"):
            js_score += 1

        for pattern, score in js_indicators:
            if (matches := _rx.compile(pattern, _rx.IGNORECASE).findall(code)):
                js_score += len(matches) * score

        return js_score >= 2.0



"""
This module provides the `rgba`, `hsla` and `hexa` classes, which offer
methods to manipulate colors in their respective color spaces.<br>

This module also provides the `Color` class, which
includes methods to work with colors in various formats.
"""

from .base.types import AnyRgba, AnyHsla, AnyHexa, Rgba, Hsla, Hexa
from .regex import Regex

from typing import Iterator, Optional, Literal, cast
import re as _re


class rgba:
    """An RGB/RGBA color object that includes a bunch of methods to manipulate the color.\n
    ----------------------------------------------------------------------------------------
    - `r` -⠀the red channel in range [0, 255] inclusive
    - `g` -⠀the green channel in range [0, 255] inclusive
    - `b` -⠀the blue channel in range [0, 255] inclusive
    - `a` -⠀the alpha channel in range [0.0, 1.0] inclusive
      or `None` if the color has no alpha channel\n
    ----------------------------------------------------------------------------------------
    Includes methods:
    - `to_hsla()` to convert to HSL color
    - `to_hexa()` to convert to HEX color
    - `has_alpha()` to check if the color has an alpha channel
    - `lighten(amount)` to create a lighter version of the color
    - `darken(amount)` to create a darker version of the color
    - `saturate(amount)` to increase color saturation
    - `desaturate(amount)` to decrease color saturation
    - `rotate(degrees)` to rotate the hue by degrees
    - `invert()` to get the inverse color
    - `grayscale()` to convert to grayscale
    - `blend(other, ratio)` to blend with another color
    - `is_dark()` to check if the color is considered dark
    - `is_light()` to check if the color is considered light
    - `is_grayscale()` to check if the color is grayscale
    - `is_opaque()` to check if the color has no transparency
    - `with_alpha(alpha)` to create a new color with different alpha
    - `complementary()` to get the complementary color"""

    def __init__(self, r: int, g: int, b: int, a: Optional[float] = None, _validate: bool = True):
        self.r: int
        """The red channel in range [0, 255] inclusive."""
        self.g: int
        """The green channel in range [0, 255] inclusive."""
        self.b: int
        """The blue channel in range [0, 255] inclusive."""
        self.a: Optional[float]
        """The alpha channel in range [0.0, 1.0] inclusive or `None` if not set."""

        if not _validate:
            self.r, self.g, self.b, self.a = r, g, b, a
            return

        if not all((0 <= x <= 255) for x in (r, g, b)):
            raise ValueError(
                f"The 'r', 'g' and 'b' parameters must be integers in range [0, 255] inclusive, got {r=} {g=} {b=}"
            )
        if a is not None and not (0.0 <= a <= 1.0):
            raise ValueError(f"The 'a' parameter must be in range [0.0, 1.0] inclusive, got {a!r}")

        self.r, self.g, self.b = r, g, b
        self.a = None if a is None else (1.0 if a > 1.0 else float(a))

    def __len__(self) -> int:
        """The number of components in the color (3 or 4)."""
        return 3 if self.a is None else 4

    def __iter__(self) -> Iterator:
        return iter((self.r, self.g, self.b) + (() if self.a is None else (self.a, )))

    def __getitem__(self, index: int) -> int | float:
        return ((self.r, self.g, self.b) + (() if self.a is None else (self.a, )))[index]

    def __eq__(self, other: object) -> bool:
        """Check if two `rgba` objects are the same color."""
        if not isinstance(other, rgba):
            return False
        return (self.r, self.g, self.b, self.a) == (other.r, other.g, other.b, other.a)

    def __ne__(self, other: object) -> bool:
        """Check if two `rgba` objects are different colors."""
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"rgba({self.r}, {self.g}, {self.b}{'' if self.a is None else f', {self.a}'})"

    def __str__(self) -> str:
        return self.__repr__()

    def dict(self) -> dict:
        """Returns the color components as a dictionary with keys `"r"`, `"g"`, `"b"` and optionally `"a"`."""
        return dict(r=self.r, g=self.g, b=self.b) if self.a is None else dict(r=self.r, g=self.g, b=self.b, a=self.a)

    def values(self) -> tuple:
        """Returns the color components as separate values `r, g, b, a`."""
        return self.r, self.g, self.b, self.a

    def to_hsla(self) -> "hsla":
        """Returns the color as `hsla()` color object."""
        h, s, l = self._rgb_to_hsl(self.r, self.g, self.b)
        return hsla(h, s, l, self.a, _validate=False)

    def to_hexa(self) -> "hexa":
        """Returns the color as `hexa()` color object."""
        return hexa("", self.r, self.g, self.b, self.a)

    def has_alpha(self) -> bool:
        """Returns `True` if the color has an alpha channel and `False` otherwise."""
        return self.a is not None

    def lighten(self, amount: float) -> "rgba":
        """Increases the colors lightness by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.r, self.g, self.b, self.a = self.to_hsla().lighten(amount).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def darken(self, amount: float) -> "rgba":
        """Decreases the colors lightness by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.r, self.g, self.b, self.a = self.to_hsla().darken(amount).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def saturate(self, amount: float) -> "rgba":
        """Increases the colors saturation by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.r, self.g, self.b, self.a = self.to_hsla().saturate(amount).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def desaturate(self, amount: float) -> "rgba":
        """Decreases the colors saturation by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.r, self.g, self.b, self.a = self.to_hsla().desaturate(amount).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def rotate(self, degrees: int) -> "rgba":
        """Rotates the colors hue by the specified number of degrees."""
        if not isinstance(degrees, int):
            raise TypeError(f"The 'degrees' parameter must be an integer, got {type(degrees)}")

        self.r, self.g, self.b, self.a = self.to_hsla().rotate(degrees).to_rgba().values()
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def invert(self, invert_alpha: bool = False) -> "rgba":
        """Inverts the color by rotating hue by 180 degrees and inverting lightness."""
        if not isinstance(invert_alpha, bool):
            raise TypeError(f"The 'invert_alpha' parameter must be a boolean, got {type(invert_alpha)}")

        self.r, self.g, self.b = 255 - self.r, 255 - self.g, 255 - self.b
        if invert_alpha and self.a is not None:
            self.a = 1 - self.a

        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def grayscale(self, method: Literal["wcag2", "wcag3", "simple", "bt601"] = "wcag2") -> "rgba":
        """Converts the color to grayscale using the luminance formula.\n
        ---------------------------------------------------------------------------
        - `method` -⠀the luminance calculation method to use:
          * `"wcag2"` WCAG 2.0 standard (default and most accurate for perception)
          * `"wcag3"` Draft WCAG 3.0 standard with improved coefficients
          * `"simple"` Simple arithmetic mean (less accurate)
          * `"bt601"` ITU-R BT.601 standard (older TV standard)"""
        # THE 'method' PARAM IS CHECKED IN 'Color.luminance()'
        self.r = self.g = self.b = int(Color.luminance(self.r, self.g, self.b, method=method))
        return rgba(self.r, self.g, self.b, self.a, _validate=False)

    def blend(self, other: Rgba, ratio: float = 0.5, additive_alpha: bool = False) -> "rgba":
        """Blends the current color with another color using the specified ratio in range [0.0, 1.0] inclusive.\n
        ----------------------------------------------------------------------------------------------------------
        - `other` -⠀the other RGBA color to blend with
        - `ratio` -⠀the blend ratio between the two colors:
          * if `ratio` is `0.0` it means 100% of the current color and 0% of the `other` color (2:0 mixture)
          * if `ratio` is `0.5` it means 50% of both colors (1:1 mixture)
          * if `ratio` is `1.0` it means 0% of the current color and 100% of the `other` color (0:2 mixture)
        - `additive_alpha` -⠀whether to blend the alpha channels additively or not"""
        if not isinstance(other, rgba):
            if Color.is_valid_rgba(other):
                other = Color.to_rgba(other)
            else:
                raise TypeError(f"The 'other' parameter must be a valid RGBA color, got {type(other)}")
        if not isinstance(ratio, float):
            raise TypeError(f"The 'ratio' parameter must be a float, got {type(ratio)}")
        elif not (0.0 <= ratio <= 1.0):
            raise ValueError(f"The 'ratio' parameter must be in range [0.0, 1.0] inclusive, got {ratio!r}")
        if not isinstance(additive_alpha, bool):
            raise TypeError(f"The 'additive_alpha' parameter must be a boolean, got {type(additive_alpha)}")

        ratio *= 2
        self.r = int(max(0, min(255, int(round((self.r * (2 - ratio)) + (other.r * ratio))))))
        self.g = int(max(0, min(255, int(round((self.g * (2 - ratio)) + (other.g * ratio))))))
        self.b = int(max(0, min(255, int(round((self.b * (2 - ratio)) + (other.b * ratio))))))
        none_alpha = self.a is None and (len(other) <= 3 or other[3] is None)

        if not none_alpha:
            self_a = 1 if self.a is None else self.a
            other_a = (other[3] if other[3] is not None else 1) if len(other) > 3 else 1

            if additive_alpha:
                self.a = max(0, min(1, (self_a * (2 - ratio)) + (other_a * ratio)))
            else:
                self.a = max(0, min(1, (self_a * (1 - (ratio / 2))) + (other_a * (ratio / 2))))

        else:
            self.a = None

        return rgba(self.r, self.g, self.b, None if none_alpha else self.a, _validate=False)

    def is_dark(self) -> bool:
        """Returns `True` if the color is considered dark (`lightness < 50%`)."""
        return self.to_hsla().is_dark()

    def is_light(self) -> bool:
        """Returns `True` if the color is considered light (`lightness >= 50%`)."""
        return not self.is_dark()

    def is_grayscale(self) -> bool:
        """Returns `True` if the color is grayscale."""
        return self.r == self.g == self.b

    def is_opaque(self) -> bool:
        """Returns `True` if the color has no transparency."""
        return self.a == 1 or self.a is None

    def with_alpha(self, alpha: float) -> "rgba":
        """Returns a new color with the specified alpha value."""
        if not isinstance(alpha, float):
            raise TypeError(f"The 'alpha' parameter must be a float, got {type(alpha)}")
        elif not (0.0 <= alpha <= 1.0):
            raise ValueError(f"The 'alpha' parameter must be in range [0.0, 1.0] inclusive, got {alpha!r}")

        return rgba(self.r, self.g, self.b, alpha, _validate=False)

    def complementary(self) -> "rgba":
        """Returns the complementary color (180 degrees on the color wheel)."""
        return self.to_hsla().complementary().to_rgba()

    @staticmethod
    def _rgb_to_hsl(r: int, g: int, b: int) -> tuple:
        """Internal method to convert RGB to HSL color space."""
        _r, _g, _b = r / 255.0, g / 255.0, b / 255.0
        max_c, min_c = max(_r, _g, _b), min(_r, _g, _b)
        l = (max_c + min_c) / 2

        if max_c == min_c:
            h = s = 0.0
        else:
            delta = max_c - min_c
            s = delta / (1 - abs(2 * l - 1))

            if max_c == _r:
                h = ((_g - _b) / delta) % 6
            elif max_c == _g:
                h = ((_b - _r) / delta) + 2
            else:
                h = ((_r - _g) / delta) + 4
            h /= 6

        return int(round(h * 360)), int(round(s * 100)), int(round(l * 100))


class hsla:
    """A HSL/HSLA color object that includes a bunch of methods to manipulate the color.\n
    ---------------------------------------------------------------------------------------
    - `h` -⠀the hue channel in range [0, 360] inclusive
    - `s` -⠀the saturation channel in range [0, 100] inclusive
    - `l` -⠀the lightness channel in range [0, 100] inclusive
    - `a` -⠀the alpha channel in range [0.0, 1.0] inclusive
      or `None` if the color has no alpha channel\n
    ---------------------------------------------------------------------------------------
    Includes methods:
    - `to_rgba()` to convert to RGB color
    - `to_hexa()` to convert to HEX color
    - `has_alpha()` to check if the color has an alpha channel
    - `lighten(amount)` to create a lighter version of the color
    - `darken(amount)` to create a darker version of the color
    - `saturate(amount)` to increase color saturation
    - `desaturate(amount)` to decrease color saturation
    - `rotate(degrees)` to rotate the hue by degrees
    - `invert()` to get the inverse color
    - `grayscale()` to convert to grayscale
    - `blend(other, ratio)` to blend with another color
    - `is_dark()` to check if the color is considered dark
    - `is_light()` to check if the color is considered light
    - `is_grayscale()` to check if the color is grayscale
    - `is_opaque()` to check if the color has no transparency
    - `with_alpha(alpha)` to create a new color with different alpha
    - `complementary()` to get the complementary color"""

    def __init__(self, h: int, s: int, l: int, a: Optional[float] = None, _validate: bool = True):
        self.h: int
        """The hue channel in range [0, 360] inclusive."""
        self.s: int
        """The saturation channel in range [0, 100] inclusive."""
        self.l: int
        """The lightness channel in range [0, 100] inclusive."""
        self.a: Optional[float]
        """The alpha channel in range [0.0, 1.0] inclusive or `None` if not set."""

        if not _validate:
            self.h, self.s, self.l, self.a = h, s, l, a
            return

        if not (0 <= h <= 360):
            raise ValueError(f"The 'h' parameter must be in range [0, 360] inclusive, got {h!r}")
        if not all((0 <= x <= 100) for x in (s, l)):
            raise ValueError(f"The 's' and 'l' parameters must be in range [0, 100] inclusive, got {s=} {l=}")
        if a is not None and not (0.0 <= a <= 1.0):
            raise ValueError(f"The 'a' parameter must be in range [0.0, 1.0] inclusive, got {a!r}")

        self.h, self.s, self.l = h, s, l
        self.a = None if a is None else (1.0 if a > 1.0 else float(a))

    def __len__(self) -> int:
        """The number of components in the color (3 or 4)."""
        return 3 if self.a is None else 4

    def __iter__(self) -> Iterator:
        return iter((self.h, self.s, self.l) + (() if self.a is None else (self.a, )))

    def __getitem__(self, index: int) -> int | float:
        return ((self.h, self.s, self.l) + (() if self.a is None else (self.a, )))[index]

    def __eq__(self, other: object) -> bool:
        """Check if two `hsla` objects are the same color."""
        if not isinstance(other, hsla):
            return False
        return (self.h, self.s, self.l, self.a) == (other.h, other.s, other.l, other.a)

    def __ne__(self, other: object) -> bool:
        """Check if two `hsla` objects are different colors."""
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"hsla({self.h}°, {self.s}%, {self.l}%{'' if self.a is None else f', {self.a}'})"

    def __str__(self) -> str:
        return self.__repr__()

    def dict(self) -> dict:
        """Returns the color components as a dictionary with keys `"h"`, `"s"`, `"l"` and optionally `"a"`."""
        return dict(h=self.h, s=self.s, l=self.l) if self.a is None else dict(h=self.h, s=self.s, l=self.l, a=self.a)

    def values(self) -> tuple:
        """Returns the color components as separate values `h, s, l, a`."""
        return self.h, self.s, self.l, self.a

    def to_rgba(self) -> "rgba":
        """Returns the color as `rgba()` color object."""
        r, g, b = self._hsl_to_rgb(self.h, self.s, self.l)
        return rgba(r, g, b, self.a, _validate=False)

    def to_hexa(self) -> "hexa":
        """Returns the color as `hexa()` color object."""
        r, g, b = self._hsl_to_rgb(self.h, self.s, self.l)
        return hexa("", r, g, b, self.a)

    def has_alpha(self) -> bool:
        """Returns `True` if the color has an alpha channel and `False` otherwise."""
        return self.a is not None

    def lighten(self, amount: float) -> "hsla":
        """Increases the colors lightness by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.l = int(min(100, self.l + (100 - self.l) * amount))
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def darken(self, amount: float) -> "hsla":
        """Decreases the colors lightness by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.l = int(max(0, self.l * (1 - amount)))
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def saturate(self, amount: float) -> "hsla":
        """Increases the colors saturation by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.s = int(min(100, self.s + (100 - self.s) * amount))
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def desaturate(self, amount: float) -> "hsla":
        """Decreases the colors saturation by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.s = int(max(0, self.s * (1 - amount)))
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def rotate(self, degrees: int) -> "hsla":
        """Rotates the colors hue by the specified number of degrees."""
        if not isinstance(degrees, int):
            raise TypeError(f"The 'degrees' parameter must be an integer, got {type(degrees)}")

        self.h = (self.h + degrees) % 360
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def invert(self, invert_alpha: bool = False) -> "hsla":
        """Inverts the color by rotating hue by 180 degrees and inverting lightness."""
        if not isinstance(invert_alpha, bool):
            raise TypeError(f"The 'invert_alpha' parameter must be a boolean, got {type(invert_alpha)}")

        self.h = (self.h + 180) % 360
        self.l = 100 - self.l
        if invert_alpha and self.a is not None:
            self.a = 1 - self.a

        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def grayscale(self, method: Literal["wcag2", "wcag3", "simple", "bt601"] = "wcag2") -> "hsla":
        """Converts the color to grayscale using the luminance formula.\n
        ---------------------------------------------------------------------------
        - `method` -⠀the luminance calculation method to use:
          * `"wcag2"` WCAG 2.0 standard (default and most accurate for perception)
          * `"wcag3"` Draft WCAG 3.0 standard with improved coefficients
          * `"simple"` Simple arithmetic mean (less accurate)
          * `"bt601"` ITU-R BT.601 standard (older TV standard)"""
        # THE 'method' PARAM IS CHECKED IN 'Color.luminance()'
        r, g, b = self._hsl_to_rgb(self.h, self.s, self.l)
        l = int(Color.luminance(r, g, b, output_type=None, method=method))
        self.h, self.s, self.l, _ = rgba(l, l, l, _validate=False).to_hsla().values()
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def blend(self, other: Hsla, ratio: float = 0.5, additive_alpha: bool = False) -> "hsla":
        """Blends the current color with another color using the specified ratio in range [0.0, 1.0] inclusive.\n
        ----------------------------------------------------------------------------------------------------------
        - `other` -⠀the other HSLA color to blend with
        - `ratio` -⠀the blend ratio between the two colors:
          * if `ratio` is `0.0` it means 100% of the current color and 0% of the `other` color (2:0 mixture)
          * if `ratio` is `0.5` it means 50% of both colors (1:1 mixture)
          * if `ratio` is `1.0` it means 0% of the current color and 100% of the `other` color (0:2 mixture)
        - `additive_alpha` -⠀whether to blend the alpha channels additively or not"""
        if not Color.is_valid_hsla(other):
            raise TypeError(f"The 'other' parameter must be a valid HSLA color, got {type(other)}")
        if not isinstance(ratio, float):
            raise TypeError(f"The 'ratio' parameter must be a float, got {type(ratio)}")
        elif not (0.0 <= ratio <= 1.0):
            raise ValueError(f"The 'ratio' parameter must be in range [0.0, 1.0] inclusive, got {ratio!r}")
        if not isinstance(additive_alpha, bool):
            raise TypeError(f"The 'additive_alpha' parameter must be a boolean, got {type(additive_alpha)}")

        self.h, self.s, self.l, self.a = self.to_rgba().blend(Color.to_rgba(other), ratio, additive_alpha).to_hsla().values()
        return hsla(self.h, self.s, self.l, self.a, _validate=False)

    def is_dark(self) -> bool:
        """Returns `True` if the color is considered dark (`lightness < 50%`)."""
        return self.l < 50

    def is_light(self) -> bool:
        """Returns `True` if the color is considered light (`lightness >= 50%`)."""
        return not self.is_dark()

    def is_grayscale(self) -> bool:
        """Returns `True` if the color is considered grayscale."""
        return self.s == 0

    def is_opaque(self) -> bool:
        """Returns `True` if the color has no transparency."""
        return self.a == 1 or self.a is None

    def with_alpha(self, alpha: float) -> "hsla":
        """Returns a new color with the specified alpha value."""
        if not isinstance(alpha, float):
            raise TypeError(f"The 'alpha' parameter must be a float, got {type(alpha)}")
        elif not (0.0 <= alpha <= 1.0):
            raise ValueError(f"The 'alpha' parameter must be in range [0.0, 1.0] inclusive, got {alpha!r}")

        return hsla(self.h, self.s, self.l, alpha, _validate=False)

    def complementary(self) -> "hsla":
        """Returns the complementary color (180 degrees on the color wheel)."""
        return hsla((self.h + 180) % 360, self.s, self.l, self.a, _validate=False)

    @classmethod
    def _hsl_to_rgb(cls, h: int, s: int, l: int) -> tuple:
        """Internal method to convert HSL to RGB color space."""
        _h, _s, _l = h / 360, s / 100, l / 100

        if _s == 0:
            r = g = b = int(_l * 255)
        else:
            q = _l * (1 + _s) if _l < 0.5 else _l + _s - _l * _s
            p = 2 * _l - q
            r = int(round(cls._hue_to_rgb(p, q, _h + 1 / 3) * 255))
            g = int(round(cls._hue_to_rgb(p, q, _h) * 255))
            b = int(round(cls._hue_to_rgb(p, q, _h - 1 / 3) * 255))

        return r, g, b

    @staticmethod
    def _hue_to_rgb(p: float, q: float, t: float) -> float:
        if t < 0:
            t += 1
        if t > 1:
            t -= 1
        if t < 1 / 6:
            return p + (q - p) * 6 * t
        if t < 1 / 2:
            return q
        if t < 2 / 3:
            return p + (q - p) * (2 / 3 - t) * 6
        return p


class hexa:
    """A HEXA color object that includes a bunch of methods to manipulate the color.\n
    --------------------------------------------------------------------------------------------
    - `color` -⠀the HEXA color string (prefix optional) or HEX integer, that can be in formats:
      * `RGB` short format without alpha (only for strings)
      * `RGBA` short format with alpha (only for strings)
      * `RRGGBB` long format without alpha (for strings and HEX integers)
      * `RRGGBBAA` long format with alpha (for strings and HEX integers)
    --------------------------------------------------------------------------------------------
    Includes methods:
    - `to_rgba()` to convert to RGB color
    - `to_hsla()` to convert to HSL color
    - `has_alpha()` to check if the color has an alpha channel
    - `lighten(amount)` to create a lighter version of the color
    - `darken(amount)` to create a darker version of the color
    - `saturate(amount)` to increase color saturation
    - `desaturate(amount)` to decrease color saturation
    - `rotate(degrees)` to rotate the hue by degrees
    - `invert()` to get the inverse color
    - `grayscale()` to convert to grayscale
    - `blend(other, ratio)` to blend with another color
    - `is_dark()` to check if the color is considered dark
    - `is_light()` to check if the color is considered light
    - `is_grayscale()` to check if the color is grayscale
    - `is_opaque()` to check if the color has no transparency
    - `with_alpha(alpha)` to create a new color with different alpha
    - `complementary()` to get the complementary color"""

    def __init__(
        self,
        color: str | int,
        _r: Optional[int] = None,
        _g: Optional[int] = None,
        _b: Optional[int] = None,
        _a: Optional[float] = None,
    ):
        self.r: int
        """The red channel in range [0, 255] inclusive."""
        self.g: int
        """The green channel in range [0, 255] inclusive."""
        self.b: int
        """The blue channel in range [0, 255] inclusive."""
        self.a: Optional[float]
        """The alpha channel in range [0.0, 1.0] inclusive or `None` if not set."""

        if all(x is not None for x in (_r, _g, _b)):
            self.r, self.g, self.b, self.a = cast(int, _r), cast(int, _g), cast(int, _b), _a
            return

        if isinstance(color, hexa):
            raise ValueError("Color is already a hexa() color object.")

        elif isinstance(color, str):
            if color.startswith("#"):
                color = color[1:].upper()
            elif color.startswith("0x"):
                color = color[2:].upper()

            if len(color) == 3:  # RGB
                self.r, self.g, self.b, self.a = (
                    int(color[0] * 2, 16),
                    int(color[1] * 2, 16),
                    int(color[2] * 2, 16),
                    None,
                )
            elif len(color) == 4:  # RGBA
                self.r, self.g, self.b, self.a = (
                    int(color[0] * 2, 16),
                    int(color[1] * 2, 16),
                    int(color[2] * 2, 16),
                    int(color[3] * 2, 16) / 255.0,
                )
            elif len(color) == 6:  # RRGGBB
                self.r, self.g, self.b, self.a = (
                    int(color[0:2], 16),
                    int(color[2:4], 16),
                    int(color[4:6], 16),
                    None,
                )
            elif len(color) == 8:  # RRGGBBAA
                self.r, self.g, self.b, self.a = (
                    int(color[0:2], 16),
                    int(color[2:4], 16),
                    int(color[4:6], 16),
                    int(color[6:8], 16) / 255.0,
                )
            else:
                raise ValueError(f"Invalid HEXA color string '{color}'. Must be in formats RGB, RGBA, RRGGBB or RRGGBBAA.")

        elif isinstance(color, int):
            self.r, self.g, self.b, self.a = Color.hex_int_to_rgba(color).values()
        else:
            raise TypeError(f"The 'color' parameter must be a string or integer, got {type(color)}")

    def __len__(self) -> int:
        """The number of components in the color (3 or 4)."""
        return 3 if self.a is None else 4

    def __iter__(self) -> Iterator:
        return iter((f"{self.r:02X}", f"{self.g:02X}", f"{self.b:02X}")
                    + (() if self.a is None else (f"{int(self.a * 255):02X}", )))

    def __getitem__(self, index: int) -> str | int:
        return ((f"{self.r:02X}", f"{self.g:02X}", f"{self.b:02X}") \
                + (() if self.a is None else (f"{int(self.a * 255):02X}", )))[index]

    def __eq__(self, other: object) -> bool:
        """Check if two `hexa` objects are the same color."""
        if not isinstance(other, hexa):
            return False
        return (self.r, self.g, self.b, self.a) == (other.r, other.g, other.b, other.a)

    def __ne__(self, other: object) -> bool:
        """Check if two `hexa` objects are different colors."""
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"hexa(#{self.r:02X}{self.g:02X}{self.b:02X}{'' if self.a is None else f'{int(self.a * 255):02X}'})"

    def __str__(self) -> str:
        return f"#{self.r:02X}{self.g:02X}{self.b:02X}{'' if self.a is None else f'{int(self.a * 255):02X}'}"

    def dict(self) -> dict:
        """Returns the color components as a dictionary with hex string values for keys `"r"`, `"g"`, `"b"` and optionally `"a"`."""
        return (
            dict(r=f"{self.r:02X}", g=f"{self.g:02X}", b=f"{self.b:02X}") if self.a is None else dict(
                r=f"{self.r:02X}",
                g=f"{self.g:02X}",
                b=f"{self.b:02X}",
                a=f"{int(self.a * 255):02X}",
            )
        )

    def values(self, round_alpha: bool = True) -> tuple:
        """Returns the color components as separate values `r, g, b, a`."""
        if not isinstance(round_alpha, bool):
            raise TypeError(f"The 'round_alpha' parameter must be a boolean, got {type(round_alpha)}")

        return self.r, self.g, self.b, None if self.a is None else (round(self.a, 2) if round_alpha else self.a)

    def to_rgba(self, round_alpha: bool = True) -> "rgba":
        """Returns the color as `rgba()` color object."""
        if not isinstance(round_alpha, bool):
            raise TypeError(f"The 'round_alpha' parameter must be a boolean, got {type(round_alpha)}")

        return rgba(
            self.r,
            self.g,
            self.b,
            None if self.a is None else (round(self.a, 2) if round_alpha else self.a),
            _validate=False,
        )

    def to_hsla(self, round_alpha: bool = True) -> "hsla":
        """Returns the color as `hsla()` color object."""
        if not isinstance(round_alpha, bool):
            raise TypeError(f"The 'round_alpha' parameter must be a boolean, got {type(round_alpha)}")

        return self.to_rgba(round_alpha).to_hsla()

    def has_alpha(self) -> bool:
        """Returns `True` if the color has an alpha channel and `False` otherwise."""
        return self.a is not None

    def lighten(self, amount: float) -> "hexa":
        """Increases the colors lightness by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.r, self.g, self.b, self.a = self.to_rgba(False).lighten(amount).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def darken(self, amount: float) -> "hexa":
        """Decreases the colors lightness by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.r, self.g, self.b, self.a = self.to_rgba(False).darken(amount).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def saturate(self, amount: float) -> "hexa":
        """Increases the colors saturation by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.r, self.g, self.b, self.a = self.to_rgba(False).saturate(amount).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def desaturate(self, amount: float) -> "hexa":
        """Decreases the colors saturation by the specified amount in range [0.0, 1.0] inclusive."""
        if not isinstance(amount, float):
            raise TypeError(f"The 'amount' parameter must be a float, got {type(amount)}")
        elif not (0.0 <= amount <= 1.0):
            raise ValueError(f"The 'amount' parameter must be in range [0.0, 1.0] inclusive, got {amount!r}")

        self.r, self.g, self.b, self.a = self.to_rgba(False).desaturate(amount).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def rotate(self, degrees: int) -> "hexa":
        """Rotates the colors hue by the specified number of degrees."""
        if not isinstance(degrees, int):
            raise TypeError(f"The 'degrees' parameter must be an integer, got {type(degrees)}")

        self.r, self.g, self.b, self.a = self.to_rgba(False).rotate(degrees).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def invert(self, invert_alpha: bool = False) -> "hexa":
        """Inverts the color by rotating hue by 180 degrees and inverting lightness."""
        if not isinstance(invert_alpha, bool):
            raise TypeError(f"The 'invert_alpha' parameter must be a boolean, got {type(invert_alpha)}")

        self.r, self.g, self.b, self.a = self.to_rgba(False).invert().values()
        if invert_alpha and self.a is not None:
            self.a = 1 - self.a

        return hexa("", self.r, self.g, self.b, self.a)

    def grayscale(self, method: Literal["wcag2", "wcag3", "simple", "bt601"] = "wcag2") -> "hexa":
        """Converts the color to grayscale using the luminance formula.\n
        ---------------------------------------------------------------------------
        - `method` -⠀the luminance calculation method to use:
          * `"wcag2"` WCAG 2.0 standard (default and most accurate for perception)
          * `"wcag3"` Draft WCAG 3.0 standard with improved coefficients
          * `"simple"` Simple arithmetic mean (less accurate)
          * `"bt601"` ITU-R BT.601 standard (older TV standard)"""
        # THE 'method' PARAM IS CHECKED IN 'Color.luminance()'
        self.r = self.g = self.b = int(Color.luminance(self.r, self.g, self.b, method=method))
        return hexa("", self.r, self.g, self.b, self.a)

    def blend(self, other: Hexa, ratio: float = 0.5, additive_alpha: bool = False) -> "hexa":
        """Blends the current color with another color using the specified ratio in range [0.0, 1.0] inclusive.\n
        ----------------------------------------------------------------------------------------------------------
        - `other` -⠀the other HEXA color to blend with
        - `ratio` -⠀the blend ratio between the two colors:
          * if `ratio` is `0.0` it means 100% of the current color and 0% of the `other` color (2:0 mixture)
          * if `ratio` is `0.5` it means 50% of both colors (1:1 mixture)
          * if `ratio` is `1.0` it means 0% of the current color and 100% of the `other` color (0:2 mixture)
        - `additive_alpha` -⠀whether to blend the alpha channels additively or not"""
        if not Color.is_valid_hexa(other):
            raise TypeError(f"The 'other' parameter must be a valid HEXA color, got {type(other)}")
        if not isinstance(ratio, float):
            raise TypeError(f"The 'ratio' parameter must be a float, got {type(ratio)}")
        elif not (0.0 <= ratio <= 1.0):
            raise ValueError(f"The 'ratio' parameter must be in range [0.0, 1.0] inclusive, got {ratio!r}")
        if not isinstance(additive_alpha, bool):
            raise TypeError(f"The 'additive_alpha' parameter must be a boolean, got {type(additive_alpha)}")

        self.r, self.g, self.b, self.a = self.to_rgba(False).blend(Color.to_rgba(other), ratio, additive_alpha).values()
        return hexa("", self.r, self.g, self.b, self.a)

    def is_dark(self) -> bool:
        """Returns `True` if the color is considered dark (`lightness < 50%`)."""
        return self.to_hsla(False).is_dark()

    def is_light(self) -> bool:
        """Returns `True` if the color is considered light (`lightness >= 50%`)."""
        return not self.is_dark()

    def is_grayscale(self) -> bool:
        """Returns `True` if the color is grayscale (`saturation == 0`)."""
        return self.to_hsla(False).is_grayscale()

    def is_opaque(self) -> bool:
        """Returns `True` if the color has no transparency (`alpha == 1.0`)."""
        return self.a == 1 or self.a is None

    def with_alpha(self, alpha: float) -> "hexa":
        """Returns a new color with the specified alpha value."""
        if not isinstance(alpha, float):
            raise TypeError(f"The 'alpha' parameter must be a float, got {type(alpha)}")
        elif not (0.0 <= alpha <= 1.0):
            raise ValueError(f"The 'alpha' parameter must be in range [0.0, 1.0] inclusive, got {alpha!r}")

        return hexa("", self.r, self.g, self.b, alpha)

    def complementary(self) -> "hexa":
        """Returns the complementary color (180 degrees on the color wheel)."""
        return self.to_hsla(False).complementary().to_hexa()


class Color:
    """This class includes methods to work with colors in different formats."""

    @classmethod
    def is_valid_rgba(cls, color: AnyRgba, allow_alpha: bool = True) -> bool:
        """Check if the given color is a valid RGBA color.\n
        -----------------------------------------------------------------
        - `color` -⠀the color to check (can be in any supported format)
        - `allow_alpha` -⠀whether to allow alpha channel in the color"""
        if not isinstance(allow_alpha, bool):
            raise TypeError(f"The 'new_tab_size' parameter must be an boolean, got {type(allow_alpha)}")

        try:
            if isinstance(color, rgba):
                return True

            elif isinstance(color, (list, tuple)):
                if allow_alpha and cls.has_alpha(color):
                    return (
                        0 <= color[0] <= 255 and 0 <= color[1] <= 255 and 0 <= color[2] <= 255
                        and (0 <= color[3] <= 1 or color[3] is None)
                    )
                elif len(color) == 3:
                    return 0 <= color[0] <= 255 and 0 <= color[1] <= 255 and 0 <= color[2] <= 255
                else:
                    return False

            elif isinstance(color, dict):
                if allow_alpha and cls.has_alpha(color):
                    return (
                        0 <= color["r"] <= 255 and 0 <= color["g"] <= 255 and 0 <= color["b"] <= 255
                        and (0 <= color["a"] <= 1 or color["a"] is None)
                    )
                elif len(color) == 3:
                    return 0 <= color["r"] <= 255 and 0 <= color["g"] <= 255 and 0 <= color["b"] <= 255
                else:
                    return False

            elif isinstance(color, str):
                return bool(_re.fullmatch(Regex.rgba_str(fix_sep=None, allow_alpha=allow_alpha), color))

        except Exception:
            pass
        return False

    @classmethod
    def is_valid_hsla(cls, color: AnyHsla, allow_alpha: bool = True) -> bool:
        """Check if the given color is a valid HSLA color.\n
        -----------------------------------------------------------------
        - `color` -⠀the color to check (can be in any supported format)
        - `allow_alpha` -⠀whether to allow alpha channel in the color"""
        try:
            if isinstance(color, hsla):
                return True

            elif isinstance(color, (list, tuple)):
                if allow_alpha and cls.has_alpha(color):
                    return (
                        0 <= color[0] <= 360 and 0 <= color[1] <= 100 and 0 <= color[2] <= 100
                        and (0 <= color[3] <= 1 or color[3] is None)
                    )
                elif len(color) == 3:
                    return 0 <= color[0] <= 360 and 0 <= color[1] <= 100 and 0 <= color[2] <= 100
                else:
                    return False

            elif isinstance(color, dict):
                if allow_alpha and cls.has_alpha(color):
                    return (
                        0 <= color["h"] <= 360 and 0 <= color["s"] <= 100 and 0 <= color["l"] <= 100
                        and (0 <= color["a"] <= 1 or color["a"] is None)
                    )
                elif len(color) == 3:
                    return 0 <= color["h"] <= 360 and 0 <= color["s"] <= 100 and 0 <= color["l"] <= 100
                else:
                    return False

            elif isinstance(color, str):
                return bool(_re.fullmatch(Regex.hsla_str(fix_sep=None, allow_alpha=allow_alpha), color))

        except Exception:
            pass
        return False

    @classmethod
    def is_valid_hexa(
        cls,
        color: AnyHexa,
        allow_alpha: bool = True,
        get_prefix: bool = False,
    ) -> bool | tuple[bool, Optional[Literal["#", "0x"]]]:
        """Check if the given color is a valid HEXA color.\n
        ---------------------------------------------------------------------------------------------------
        - `color` -⠀the color to check (can be in any supported format)
        - `allow_alpha` -⠀whether to allow alpha channel in the color
        - `get_prefix` -⠀if true, the prefix used in the color (if any) is returned along with validity"""
        try:
            if isinstance(color, hexa):
                return (True, "#") if get_prefix else True

            elif isinstance(color, int):
                is_valid = 0x000000 <= color <= (0xFFFFFFFF if allow_alpha else 0xFFFFFF)
                return (is_valid, "0x") if get_prefix else is_valid

            elif isinstance(color, str):
                prefix: Optional[Literal["#", "0x"]]
                color, prefix = ((color[1:], "#") if color.startswith("#") else
                                 (color[2:], "0x") if color.startswith("0x") else (color, None))
                return (
                    (bool(_re.fullmatch(Regex.hexa_str(allow_alpha=allow_alpha), color)), prefix) \
                    if get_prefix else bool(_re.fullmatch(Regex.hexa_str(allow_alpha=allow_alpha), color))
                )

        except Exception:
            pass
        return (False, None) if get_prefix else False

    @classmethod
    def is_valid(cls, color: AnyRgba | AnyHsla | AnyHexa, allow_alpha: bool = True) -> bool:
        """Check if the given color is a valid RGBA, HSLA or HEXA color.\n
        -------------------------------------------------------------------
        - `color` -⠀the color to check (can be in any supported format)
        - `allow_alpha` -⠀whether to allow alpha channel in the color"""
        return bool(
            cls.is_valid_rgba(color, allow_alpha) \
            or cls.is_valid_hsla(color, allow_alpha) \
            or cls.is_valid_hexa(color, allow_alpha)
        )

    @classmethod
    def has_alpha(cls, color: Rgba | Hsla | Hexa) -> bool:
        """Check if the given color has an alpha channel.\n
        ---------------------------------------------------------------------------
        - `color` -⠀the color to check (can be in any supported format)"""
        if isinstance(color, (rgba, hsla, hexa)):
            return color.has_alpha()

        if cls.is_valid_hexa(color):
            if isinstance(color, str):
                if color.startswith("#"):
                    color = color[1:]
                elif color.startswith("0x"):
                    color = color[2:]
                return len(color) == 4 or len(color) == 8
            if isinstance(color, int):
                hex_length = len(f"{color:X}")
                return hex_length == 4 or hex_length == 8

        elif isinstance(color, str):
            if parsed_rgba := cls.str_to_rgba(color, only_first=True):
                return cast(rgba, parsed_rgba).has_alpha()
            if parsed_hsla := cls.str_to_hsla(color, only_first=True):
                return cast(hsla, parsed_hsla).has_alpha()

        elif isinstance(color, (list, tuple)) and len(color) == 4 and color[3] is not None:
            return True
        elif isinstance(color, dict) and len(color) == 4 and color["a"] is not None:
            return True

        return False

    @classmethod
    def to_rgba(cls, color: Rgba | Hsla | Hexa) -> rgba:
        """Will try to convert any color type to a color of type RGBA.\n
        ---------------------------------------------------------------------
        - `color` -⠀the color to convert (can be in any supported format)"""
        if isinstance(color, (hsla, hexa)):
            return color.to_rgba()
        elif cls.is_valid_hsla(color):
            return cls._parse_hsla(color).to_rgba()
        elif cls.is_valid_hexa(color):
            return hexa(cast(str | int, color)).to_rgba()
        elif cls.is_valid_rgba(color):
            return cls._parse_rgba(color)
        raise ValueError(f"Could not convert color {color!r} to RGBA.")

    @classmethod
    def to_hsla(cls, color: Rgba | Hsla | Hexa) -> hsla:
        """Will try to convert any color type to a color of type HSLA.\n
        ---------------------------------------------------------------------
        - `color` -⠀the color to convert (can be in any supported format)"""
        if isinstance(color, (rgba, hexa)):
            return color.to_hsla()
        elif cls.is_valid_rgba(color):
            return cls._parse_rgba(color).to_hsla()
        elif cls.is_valid_hexa(color):
            return hexa(cast(str | int, color)).to_hsla()
        elif cls.is_valid_hsla(color):
            return cls._parse_hsla(color)
        raise ValueError(f"Could not convert color {color!r} to HSLA.")

    @classmethod
    def to_hexa(cls, color: Rgba | Hsla | Hexa) -> hexa:
        """Will try to convert any color type to a color of type HEXA.\n
        ---------------------------------------------------------------------
        - `color` -⠀the color to convert (can be in any supported format)"""
        if isinstance(color, (rgba, hsla)):
            return color.to_hexa()
        elif cls.is_valid_rgba(color):
            return cls._parse_rgba(color).to_hexa()
        elif cls.is_valid_hsla(color):
            return cls._parse_hsla(color).to_hexa()
        elif cls.is_valid_hexa(color):
            return color if isinstance(color, hexa) else hexa(cast(str | int, color))
        raise ValueError(f"Could not convert color {color!r} to HEXA")

    @classmethod
    def str_to_rgba(cls, string: str, only_first: bool = False) -> Optional[rgba | list[rgba]]:
        """Will try to recognize RGBA colors inside a string and output the found ones as RGBA objects.\n
        ---------------------------------------------------------------------------------------------------------------
        - `string` -⠀the string to search for RGBA colors
        - `only_first` -⠀if true, only the first found color will be returned, otherwise a list of all found colors"""
        if only_first:
            if not (match := _re.search(Regex.rgba_str(allow_alpha=True), string)):
                return None
            m = match.groups()
            return rgba(
                int(m[0]),
                int(m[1]),
                int(m[2]),
                ((int(m[3]) if "." not in m[3] else float(m[3])) if m[3] else None),
                _validate=False,
            )

        else:
            if not (matches := _re.findall(Regex.rgba_str(allow_alpha=True), string)):
                return None
            return [
                rgba(
                    int(m[0]),
                    int(m[1]),
                    int(m[2]),
                    ((int(m[3]) if "." not in m[3] else float(m[3])) if m[3] else None),
                    _validate=False,
                ) for m in matches
            ]

    @classmethod
    def str_to_hsla(cls, string: str, only_first: bool = False) -> Optional[hsla | list[hsla]]:
        """Will try to recognize HSLA colors inside a string and output the found ones as HSLA objects.\n
        ---------------------------------------------------------------------------------------------------------------
        - `string` -⠀the string to search for HSLA colors
        - `only_first` -⠀if true, only the first found color will be returned, otherwise a list of all found colors"""
        if only_first:
            if not (match := _re.search(Regex.hsla_str(allow_alpha=True), string)):
                return None
            m = match.groups()
            return hsla(
                int(m[0]),
                int(m[1]),
                int(m[2]),
                ((int(m[3]) if "." not in m[3] else float(m[3])) if m[3] else None),
                _validate=False,
            )

        else:
            if not (matches := _re.findall(Regex.hsla_str(allow_alpha=True), string)):
                return None
            return [
                hsla(
                    int(m[0]),
                    int(m[1]),
                    int(m[2]),
                    ((int(m[3]) if "." not in m[3] else float(m[3])) if m[3] else None),
                    _validate=False,
                ) for m in matches
            ]

    @classmethod
    def rgba_to_hex_int(
        cls,
        r: int,
        g: int,
        b: int,
        a: Optional[float] = None,
        preserve_original: bool = False,
    ) -> int:
        """Convert RGBA channels to a HEXA integer (alpha is optional).\n
        --------------------------------------------------------------------------------------------
        - `r`, `g`, `b` -⠀the red, green and blue channels (`0` – `255`)
        - `a` -⠀the alpha channel (`0.0` – `1.0`) or `None` if not set
        - `preserve_original` -⠀whether to preserve the original color exactly (explained below)\n
        --------------------------------------------------------------------------------------------
        To preserve leading zeros, the function will add a `1` at the beginning, if the HEX integer
        would start with a `0`.
        This could affect the color a little bit, but will make sure, that it won't be interpreted
        as a completely different color, when initializing it as a `hexa()` color or changing it
        back to RGBA using `Color.hex_int_to_rgba()`."""
        if not all((0 <= c <= 255) for c in (r, g, b)):
            raise ValueError(f"The 'r', 'g' and 'b' parameters must be integers in [0, 255], got {r=} {g=} {b=}")
        if a is not None and not (0.0 <= a <= 1.0):
            raise ValueError(f"The 'a' parameter must be a float in [0.0, 1.0] or None, got {a!r}")

        r = max(0, min(255, int(r)))
        g = max(0, min(255, int(g)))
        b = max(0, min(255, int(b)))

        if a is None:
            hex_int = (r << 16) | (g << 8) | b
            if not preserve_original and (hex_int & 0xF00000) == 0:
                hex_int |= 0x010000
        else:
            a = max(0, min(255, int(a * 255)))
            hex_int = (r << 24) | (g << 16) | (b << 8) | a
            if not preserve_original and r == 0:
                hex_int |= 0x01000000

        return hex_int

    @classmethod
    def hex_int_to_rgba(cls, hex_int: int, preserve_original: bool = False) -> rgba:
        """Convert a HEX integer to RGBA channels.\n
        -------------------------------------------------------------------------------------------
        - `hex_int` -⠀the HEX integer to convert
        - `preserve_original` -⠀whether to preserve the original color exactly (explained below)\n
        -------------------------------------------------------------------------------------------
        If the red channel is `1` after conversion, it will be set to `0`, because when converting
        from RGBA to a HEX integer, the first `0` will be set to `1` to preserve leading zeros.
        This is the correction, so the color doesn't even look slightly different."""
        if not (0 <= hex_int <= 0xFFFFFFFF):
            raise ValueError(f"Expected HEX integer in range [0x000000, 0xFFFFFFFF] inclusive, got 0x{hex_int:X}")

        if len(hex_str := f"{hex_int:X}") <= 6:
            hex_str = hex_str.zfill(6)
            return rgba(
                r if (r := int(hex_str[0:2], 16)) != 1 or preserve_original else 0,
                int(hex_str[2:4], 16),
                int(hex_str[4:6], 16),
                None,
                _validate=False,
            )

        elif len(hex_str) <= 8:
            hex_str = hex_str.zfill(8)
            return rgba(
                r if (r := int(hex_str[0:2], 16)) != 1 or preserve_original else 0,
                int(hex_str[2:4], 16),
                int(hex_str[4:6], 16),
                int(hex_str[6:8], 16) / 255.0,
                _validate=False,
            )

        else:
            raise ValueError(f"Could not convert HEX integer 0x{hex_int:X} to RGBA color.")

    @classmethod
    def luminance(
        cls,
        r: int,
        g: int,
        b: int,
        output_type: Optional[type[int | float]] = None,
        method: Literal["wcag2", "wcag3", "simple", "bt601"] = "wcag2",
    ) -> int | float:
        """Calculates the relative luminance of a color according to various standards.\n
        ----------------------------------------------------------------------------------
        - `r`, `g`, `b` -⠀the red, green and blue channels in range [0, 255] inclusive
        - `output_type` -⠀the range of the returned luminance value:
          * `int` returns integer in range [0, 100] inclusive
          * `float` returns float in range [0.0, 1.0] inclusive
          * `None` returns integer in range [0, 255] inclusive
        - `method` -⠀the luminance calculation method to use:
          * `"wcag2"` WCAG 2.0 standard (default and most accurate for perception)
          * `"wcag3"` Draft WCAG 3.0 standard with improved coefficients
          * `"simple"` Simple arithmetic mean (less accurate)
          * `"bt601"` ITU-R BT.601 standard (older TV standard)"""
        if not all(0 <= c <= 255 for c in (r, g, b)):
            raise ValueError(f"The 'r', 'g' and 'b' parameters must be integers in [0, 255], got {r=} {g=} {b=}")
        if output_type not in {int, float, None}:
            raise TypeError(f"The 'output_type' parameter must be either 'int', 'float' or 'None', got {output_type!r}")

        _r, _g, _b = r / 255.0, g / 255.0, b / 255.0

        if method == "simple":
            luminance = (_r + _g + _b) / 3
        elif method == "bt601":
            luminance = 0.299 * _r + 0.587 * _g + 0.114 * _b
        elif method == "wcag3":
            _r = cls._linearize_srgb(_r)
            _g = cls._linearize_srgb(_g)
            _b = cls._linearize_srgb(_b)
            luminance = 0.2126729 * _r + 0.7151522 * _g + 0.0721750 * _b
        else:
            _r = cls._linearize_srgb(_r)
            _g = cls._linearize_srgb(_g)
            _b = cls._linearize_srgb(_b)
            luminance = 0.2126 * _r + 0.7152 * _g + 0.0722 * _b

        if output_type == int:
            return round(luminance * 100)
        elif output_type == float:
            return luminance
        else:
            return round(luminance * 255)

    @classmethod
    def text_color_for_on_bg(cls, text_bg_color: Rgba | Hexa) -> rgba | hexa | int:
        """Returns either black or white text color for optimal contrast on the given background color.\n
        --------------------------------------------------------------------------------------------------
        - `text_bg_color` -⠀the background color (can be in RGBA or HEXA format)"""
        was_hexa, was_int = cls.is_valid_hexa(text_bg_color), isinstance(text_bg_color, int)

        text_bg_color = cls.to_rgba(text_bg_color)
        brightness = 0.2126 * text_bg_color[0] + 0.7152 * text_bg_color[1] + 0.0722 * text_bg_color[2]

        return (
            (0xFFFFFF if was_int else hexa("", 255, 255, 255)) if was_hexa \
            else rgba(255, 255, 255, _validate=False)
        ) if brightness < 128 else (
            (0x000 if was_int else hexa("", 0, 0, 0)) if was_hexa \
            else rgba(0, 0, 0, _validate=False)
        )

    @classmethod
    def adjust_lightness(cls, color: Rgba | Hexa, lightness_change: float) -> rgba | hexa:
        """In- or decrease the lightness of the input color.\n
        ------------------------------------------------------------------
        - `color` -⠀the color to adjust (can be in RGBA or HEXA format)
        - `lightness_change` -⠀the amount to change the lightness by,
          in range `-1.0` (darken by 100%) and `1.0` (lighten by 100%)"""
        was_hexa = cls.is_valid_hexa(color)

        if not (-1.0 <= lightness_change <= 1.0):
            raise ValueError(
                f"The 'lightness_change' parameter must be in range [-1.0, 1.0] inclusive, got {lightness_change!r}"
            )

        hsla_color: hsla = cls.to_hsla(color)

        h, s, l, a = (
            int(hsla_color[0]), int(hsla_color[1]), int(hsla_color[2]), \
            hsla_color[3] if cls.has_alpha(hsla_color) else None
        )
        l = int(max(0, min(100, l + lightness_change * 100)))

        return (
            hsla(h, s, l, a, _validate=False).to_hexa() if was_hexa \
            else hsla(h, s, l, a, _validate=False).to_rgba()
        )

    @classmethod
    def adjust_saturation(cls, color: Rgba | Hexa, saturation_change: float) -> rgba | hexa:
        """In- or decrease the saturation of the input color.\n
        -----------------------------------------------------------------------
        - `color` -⠀the color to adjust (can be in RGBA or HEXA format)
        - `saturation_change` -⠀the amount to change the saturation by,
          in range `-1.0` (saturate by 100%) and `1.0` (desaturate by 100%)"""
        was_hexa = cls.is_valid_hexa(color)

        if not (-1.0 <= saturation_change <= 1.0):
            raise ValueError(
                f"The 'saturation_change' parameter must be in range [-1.0, 1.0] inclusive, got {saturation_change!r}"
            )

        hsla_color: hsla = cls.to_hsla(color)

        h, s, l, a = (
            int(hsla_color[0]), int(hsla_color[1]), int(hsla_color[2]), \
            hsla_color[3] if cls.has_alpha(hsla_color) else None
        )
        s = int(max(0, min(100, s + saturation_change * 100)))

        return (
            hsla(h, s, l, a, _validate=False).to_hexa() if was_hexa \
            else hsla(h, s, l, a, _validate=False).to_rgba()
        )

    @classmethod
    def _parse_rgba(cls, color: AnyRgba) -> rgba:
        """Internal method to parse a color to an RGBA object."""
        if isinstance(color, rgba):
            return color
        elif isinstance(color, (list, tuple)):
            if len(color) == 4:
                return rgba(color[0], color[1], color[2], color[3], _validate=False)
            elif len(color) == 3:
                return rgba(color[0], color[1], color[2], None, _validate=False)
        elif isinstance(color, dict):
            return rgba(color["r"], color["g"], color["b"], color.get("a"), _validate=False)
        elif isinstance(color, str):
            if parsed := cls.str_to_rgba(color, only_first=True):
                return cast(rgba, parsed)
        raise ValueError(f"Could not parse RGBA color: {color!r}")

    @classmethod
    def _parse_hsla(cls, color: AnyHsla) -> hsla:
        """Internal method to parse a color to an HSLA object."""
        if isinstance(color, hsla):
            return color
        elif isinstance(color, (list, tuple)):
            if len(color) == 4:
                return hsla(color[0], color[1], color[2], color[3], _validate=False)
            elif len(color) == 3:
                return hsla(color[0], color[1], color[2], None, _validate=False)
        elif isinstance(color, dict):
            return hsla(color["h"], color["s"], color["l"], color.get("a"), _validate=False)
        elif isinstance(color, str):
            if parsed := cls.str_to_hsla(color, only_first=True):
                return cast(hsla, parsed)
        raise ValueError(f"Could not parse HSLA color: {color!r}")

    @staticmethod
    def _linearize_srgb(c: float) -> float:
        """Helper method to linearize sRGB component following the WCAG standard."""
        if not (0.0 <= c <= 1.0):
            raise ValueError(f"The 'c' parameter must be in range [0.0, 1.0] inclusive, got {c!r}")

        if c <= 0.03928:
            return c / 12.92
        else:
            return ((c + 0.055) / 1.055)**2.4



from xulbux.env_path import EnvPath

#
################################################## EnvPath TESTS ##################################################


def test_get_paths():
    paths = EnvPath.paths()
    paths_list = EnvPath.paths(as_list=True)
    assert paths
    assert paths_list
    assert isinstance(paths, str)
    assert isinstance(paths_list, list)
    assert len(paths_list) > 0
    assert isinstance(paths_list[0], str)


def test_add_path():
    EnvPath.add_path(base_dir=True)


def test_has_path():
    assert EnvPath.has_path(base_dir=True)


def test_remove_path():
    EnvPath.remove_path(base_dir=True)
    assert not EnvPath.has_path(base_dir=True)



"""
This module provides the `Json` class, which includes methods to read,
create and update JSON files, with support for comments inside the JSON data.
"""

from .data import Data
from .file import File
from .path import Path

from typing import Literal, Any, cast
import json as _json


class Json:
    """This class provides methods to read, create and update JSON files,
    with support for comments inside the JSON data."""

    @classmethod
    def read(
        cls,
        json_file: str,
        comment_start: str = ">>",
        comment_end: str = "<<",
        return_original: bool = False,
    ) -> dict | tuple[dict, dict]:
        """Read JSON files, ignoring comments.\n
        ------------------------------------------------------------------------------------
        - `json_file` -⠀the path (relative or absolute) to the JSON file to read
        - `comment_start` -⠀the string that indicates the start of a comment
        - `comment_end` -⠀the string that indicates the end of a comment
        - `return_original` -⠀if true, the original JSON data is returned additionally:<br>
          ```python
        (processed_json, original_json)
          ```\n
        ------------------------------------------------------------------------------------
        For more detailed information about the comment handling,
        see the `Data.remove_comments()` method documentation."""
        if not json_file.endswith(".json"):
            json_file += ".json"
        if (file_path := Path.extend_or_make(json_file, prefer_script_dir=True)) is None:
            raise FileNotFoundError(f"Could not find JSON file: {json_file}")

        with open(file_path, "r") as f:
            content = f.read()

        try:
            data = _json.loads(content)
        except _json.JSONDecodeError as e:
            raise ValueError(f"Error parsing JSON in '{file_path}':  {str(e)}")

        if not (processed_data := dict(Data.remove_comments(data, comment_start, comment_end))):
            raise ValueError(f"The JSON file '{file_path}' is empty or contains only comments.")

        return (processed_data, data) if return_original else processed_data

    @classmethod
    def create(
        cls,
        json_file: str,
        data: dict,
        indent: int = 2,
        compactness: Literal[0, 1, 2] = 1,
        force: bool = False,
    ) -> str:
        """Create a nicely formatted JSON file from a dictionary.\n
        ---------------------------------------------------------------------------
        - `json_file` -⠀the path (relative or absolute) to the JSON file to create
        - `data` -⠀the dictionary data to write to the JSON file
        - `indent` -⠀the amount of spaces to use for indentation
        - `compactness` -⠀can be `0`, `1` or `2` and indicates how compact
          the data should be formatted (see `Data.render()` for more info)
        - `force` -⠀if true, will overwrite existing files
          without throwing an error (errors explained below)\n
        ---------------------------------------------------------------------------
        The method will throw a `FileExistsError` if a file with the same
        name already exists and a `SameContentFileExistsError` if a file
        with the same name and same content already exists."""
        if not json_file.endswith(".json"):
            json_file += ".json"

        File.create(
            file_path=(file_path := Path.extend_or_make(json_file, prefer_script_dir=True)),
            content=Data.render(data=data, indent=indent, compactness=compactness, as_json=True),
            force=force,
        )

        return file_path

    @classmethod
    def update(
        cls,
        json_file: str,
        update_values: dict[str, Any],
        comment_start: str = ">>",
        comment_end: str = "<<",
        path_sep: str = "->",
    ) -> None:
        """Update single/multiple values inside JSON files,
        without needing to know the rest of the data.\n
        -----------------------------------------------------------------------------------
        - `json_file` -⠀the path (relative or absolute) to the JSON file to update
        - `update_values` -⠀a dictionary with the paths to the values to update
          and the new values to set (see explanation below – section 2)
        - `comment_start` -⠀the string that indicates the start of a comment
        - `comment_end` -⠀the string that indicates the end of a comment
        - `path_sep` -⠀the separator used inside the value-paths in `update_values`\n
        -----------------------------------------------------------------------------------
        For more detailed information about the comment handling,
        see the `Data.remove_comments()` method documentation.\n
        -----------------------------------------------------------------------------------
        The `update_values` is a dictionary, where the keys are the paths
        to the data to update, and the values are the new values to set.\n
        For example for this JSON data:
        ```python
        {
            "healthy": {
                "fruits": ["apples", "bananas", "oranges"],
                "vegetables": ["carrots", "broccoli", "celery"]
            }
        }
        ```
        … the `update_values` dictionary could look like this:
        ```python
        {
            # CHANGE FIRST LIST-VALUE UNDER 'fruits' TO "strawberries"
            "healthy->fruits->0": "strawberries",
            # CHANGE VALUE OF KEY 'vegetables' TO [1, 2, 3]
            "healthy->vegetables": [1, 2, 3]
        }
        ```
        In this example, if you want to change the value of `"apples"`,
        you can use `healthy->fruits->apples` as the value-path.<br>
        If you don't know that the first list item is `"apples"`,
        you can use the items list index inside the value-path, so `healthy->fruits->0`.\n
        ⇾ If the given value-path doesn't exist, it will be created."""
        processed_data, data = cls.read(
            json_file=json_file,
            comment_start=comment_start,
            comment_end=comment_end,
            return_original=True,
        )

        update: dict[str, Any] = {}
        for val_path, new_val in update_values.items():
            try:
                if (path_id := Data.get_path_id(data=processed_data, value_paths=val_path, path_sep=path_sep)) is not None:
                    update[cast(str, path_id)] = new_val
                else:
                    data = cls._create_nested_path(data, val_path.split(path_sep), new_val)
            except Exception:
                data = cls._create_nested_path(data, val_path.split(path_sep), new_val)

        if update:
            data = Data.set_value_by_path_id(data, update)

        cls.create(json_file=json_file, data=dict(data), force=True)

    @staticmethod
    def _create_nested_path(data_obj: dict, path_keys: list[str], value: Any) -> dict:
        """Internal method that creates nested dictionaries/lists based on the
        given path keys and sets the specified value at the end of the path."""
        last_idx, current = len(path_keys) - 1, data_obj

        for i, key in enumerate(path_keys):
            if i == last_idx:
                if isinstance(current, dict):
                    current[key] = value
                elif isinstance(current, list) and key.isdigit():
                    idx = int(key)
                    while len(current) <= idx:
                        current.append(None)
                    current[idx] = value
                else:
                    raise TypeError(f"Cannot set key '{key}' on {type(current)}")

            else:
                next_key = path_keys[i + 1]
                if isinstance(current, dict):
                    if key not in current:
                        current[key] = [] if next_key.isdigit() else {}
                    current = current[key]
                elif isinstance(current, list) and key.isdigit():
                    idx = int(key)
                    while len(current) <= idx:
                        current.append(None)
                    if current[idx] is None:
                        current[idx] = [] if next_key.isdigit() else {}
                    current = current[idx]
                else:
                    raise TypeError(f"Cannot navigate through {type(current)}")

        return data_obj
